import asyncio
import itertools
import json
import math
import pytest

import numpy as np
from gameanalysis import rsgame

from egta import simsched


@pytest.fixture
def jgame():
    with open('cdasim/game.json') as f:
        return json.load(f)


@pytest.fixture
def conf(jgame):
    return jgame['configuration']


@pytest.fixture
def game(jgame):
    return rsgame.emptygame_json(jgame)


@pytest.mark.asyncio
async def test_basic_profile(conf, game):
    profs = game.random_profiles(20)
    cmd = ['python3', 'cdasim/sim.py', '--single', '1']

    async with simsched.simsched(game, conf, cmd) as sched:
        assert rsgame.emptygame_copy(sched) == game
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)


@pytest.mark.asyncio
async def test_delayed_fail(conf, game):
    prof = game.random_profile()
    cmd = ['bash', '-c', 'sleep 1 && false']

    async with simsched.simsched(game, conf, cmd) as sched:
        with pytest.raises(RuntimeError):
            await sched.sample_payoffs(prof)


@pytest.mark.asyncio
async def test_immediate_fail(conf, game):
    prof = game.random_profile()
    cmd = ['bash', '-c', 'false']

    async with simsched.simsched(game, conf, cmd) as sched:
        with pytest.raises(RuntimeError):
            await asyncio.gather(sched.sample_payoffs(prof),
                                 sched.sample_payoffs(prof))


@pytest.mark.asyncio
async def test_early_exit(conf, game):
    profs = game.random_profiles(20)
    cmd = ['bash', '-c', 'while read line; do : ; done']

    async with simsched.simsched(game, conf, cmd) as sched:
        # Schedule promises but don't await
        asyncio.ensure_future(asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs]))
        await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_read_delay_fail(conf, game):
    cmd = ['bash', '-c', 'read line && sleep 0.1 && false']

    async with simsched.simsched(game, conf, cmd) as sched:
        future = asyncio.ensure_future(sched.sample_payoffs(
            game.random_profile()))
        with pytest.raises(RuntimeError):
            await future


@pytest.mark.asyncio
async def test_read_delay_schedule_fail(conf, game):
    cmd = ['bash', '-c', 'read line && sleep 0.2 && false']

    async with simsched.simsched(game, conf, cmd) as sched:
        future = asyncio.ensure_future(sched.sample_payoffs(
            game.random_profile()))
        await asyncio.sleep(1)  # make sure process is dead
        # Since we're using asyncio, the future actually gets assigned the
        # error before this one kicks off.
        asyncio.ensure_future(sched.sample_payoffs(game.random_profile()))
        with pytest.raises(RuntimeError):
            await future


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_ignore_terminate_fail(conf, game):
    cmd = ['bash', '-c', 'trap "" SIGTERM && sleep 20']
    async with simsched.simsched(game, conf, cmd):
        # Wait for term to be captured
        await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_json_decode_fail(conf, game):
    cmd = ['bash', '-c',
           'read line && echo "[" && while read line; do :; done']
    prof = game.random_profile()
    async with simsched.simsched(game, conf, cmd) as sched:
        with pytest.raises(json.decoder.JSONDecodeError):
            await asyncio.gather(sched.sample_payoffs(prof),
                                 sched.sample_payoffs(prof),
                                 sched.sample_payoffs(prof))


@pytest.mark.asyncio
async def test_bad_command_exit(conf, game):
    with pytest.raises(FileNotFoundError):
        async with simsched.simsched(game, conf, ['unknown']):
            pass  # pragma: no cover never called


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_buffer_blocking(conf, game):
    prof = game.random_profile()
    cmd = ['python3', 'cdasim/sim.py', '--single', '1']

    async with simsched.simsched(game, conf, cmd) as sched:
        await sched.sample_payoffs(prof)
        bprof = json.dumps(sched._base, separators=(',', ':')).encode('utf8')
        # number we have to write to blow out buffer
        num = 5 * math.ceil(4096 / (len(bprof) + 1))
        await asyncio.gather(*[
            sched.sample_payoffs(p) for p in itertools.repeat(prof, num)])



import async_generator
import asyncio
import random

import numpy as np
import pytest
from egtaonline import api
from egtaonline import mockserver
from gameanalysis import rsgame

from egta import countsched
from egta import eosched


@pytest.fixture
@async_generator.async_generator
async def egta_info():
    game = rsgame.emptygame([4, 4], [11, 11])
    async with mockserver.server() as server, api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10))
        strats = dict(zip(game.role_names, game.strat_names))
        symgrps = list(zip(game.role_names, game.num_role_players,
                           game.strat_names))
        await sim.add_strategies(strats)
        egame = await egta.get_canon_game(sim['id'], symgrps)
        await async_generator.yield_((server, egta, game, egame['id']))


@pytest.mark.asyncio
async def test_basic_profile(egta_info):
    _, egta, game, game_id = egta_info
    profs = game.random_profiles(20)

    # Schedule all new profiles and verify it works
    async with eosched.eosched(
            game, egta, game_id, 0.1, 1, 10, 0, 0) as sched:
        assert game == rsgame.emptygame_copy(sched)
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
    pays = np.stack(awaited)
    assert np.allclose(pays[profs == 0], 0)

    # Schedule old profiles and verify it still works
    async with eosched.eosched(
            game, egta, game_id, 0.1, 1, 10, 0, 0) as sched:
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
    pays = np.stack(awaited)
    assert np.allclose(pays[profs == 0], 0)

    # Schedule two at a time, in two batches
    async with eosched.eosched(
            game, egta, game_id, 0.1, 2, 10, 0, 0) as base_sched:
        sched = countsched.CountScheduler(base_sched, 2)
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
    pays = np.stack(awaited)
    assert np.allclose(pays[profs == 0], 0)

    # Try again now that everything should be scheduled
    async with eosched.eosched(
            game, egta, game_id, 0.1, 2, 10, 0, 0) as base_sched:
        sched = countsched.CountScheduler(base_sched, 2)
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
    pays = np.stack(awaited)
    assert np.allclose(pays[profs == 0], 0)


@pytest.mark.asyncio
async def test_exception_in_create(egta_info):
    server, egta, game, game_id = egta_info
    server.throw_exception(TimeoutError)
    with pytest.raises(TimeoutError):
        async with eosched.eosched(
                game, egta, game_id, 0.1, 1, 25, 0, 0):
            pass  # pragma: no cover


@pytest.mark.asyncio
async def test_exception_in_get(egta_info):
    server, egta, game, game_id = egta_info
    profs = game.random_profiles(20)

    async with eosched.eosched(
            game, egta, game_id, 0.1, 1, 10, 0, 0) as sched:
        futures = asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])
        await asyncio.sleep(0.1)
        server.throw_exception(TimeoutError)
        await asyncio.sleep(0.1)
        with pytest.raises(TimeoutError):
            await futures


@pytest.mark.asyncio
async def test_exception_in_schedule(egta_info):
    server, egta, game, game_id = egta_info
    prof = game.random_profile()

    async with eosched.eosched(
            game, egta, game_id, 0.1, 1, 25, 0, 0) as sched:
        # so that enough calls to get_requirements are made
        server.throw_exception(TimeoutError)
        await asyncio.sleep(0.1)
        with pytest.raises(TimeoutError):
            await sched.sample_payoffs(prof)


@pytest.mark.asyncio
async def test_scheduler_deactivate(egta_info):
    _, egta, game, game_id = egta_info

    # Schedule all new profiles and verify it works
    # This first time should have to wait to schedule more
    async with eosched.eosched(
            game, egta, game_id, 0.1, 1, 10, 0, 0) as sched:
        # Deactivate scheduler
        async for esched in egta.get_generic_schedulers():
            await esched.deactivate()
        with pytest.raises(AssertionError):
            await sched.sample_payoffs(game.random_profile())



import numpy as np
import pytest
from gameanalysis import gamegen
from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis.reduction import deviation_preserving as dpr

from egta import asyncgame
from egta import innerloop
from egta import gamesched
from egta import schedgame


games = [
    ([1], [2]),
    ([2], [2]),
    ([2, 1], [1, 2]),
    ([1, 2], [2, 1]),
    ([2, 2], [2, 2]),
    ([3, 2], [2, 3]),
    ([1, 1, 1], [2, 2, 2]),
]


def verify_dist_thresh(eqa, thresh=0.1):
    for i, eqm in enumerate(eqa[:-1], 1):
        assert np.all(thresh ** 2 <= np.sum((eqm - eqa[i:]) ** 2, 1))


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
@pytest.mark.parametrize('_', range(1))
async def test_innerloop_simple(players, strats, _):
    sgame = gamegen.samplegame(players, strats)
    sched = gamesched.samplegamesched(sgame)
    eqa = await innerloop.inner_loop(schedgame.schedgame(sched))
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
async def test_innerloop_game(players, strats):
    game = gamegen.samplegame(players, strats)
    sched = gamesched.gamesched(game)
    eqas = await innerloop.inner_loop(schedgame.schedgame(sched))
    verify_dist_thresh(eqas)
    eqag = await innerloop.inner_loop(asyncgame.wrap(game))
    equality = np.isclose(eqas, eqag[:, None], atol=1e-3).all(2)
    assert equality.any(1).all() and equality.any(0).all()


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
async def test_innerloop_dpr(players, strats):
    redgame = rsgame.emptygame(players, strats)
    fullgame = rsgame.emptygame(redgame.num_role_players ** 2,
                                redgame.num_role_strats)
    profs = dpr.expand_profiles(fullgame, redgame.all_profiles())
    pays = np.random.random(profs.shape)
    pays[profs == 0] = 0
    sgame = paygame.samplegame_replace(fullgame, profs, [pays[:, None]])
    sched = gamesched.samplegamesched(sgame)
    game = schedgame.schedgame(sched, dpr, redgame.num_role_players)
    eqa = await innerloop.inner_loop(game)
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
async def test_innerloop_by_role_simple(players, strats):
    sgame = gamegen.samplegame(players, strats)
    sched = gamesched.samplegamesched(sgame)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched), devs_by_role=True)
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
@pytest.mark.parametrize('when', ['pre', 'post'])
@pytest.mark.parametrize('count', [1, 5, 10])
@pytest.mark.parametrize('players,strats',
                         [[[2, 3], [3, 2]], [[4, 3], [3, 4]]])
async def test_innerloop_failures(players, strats, count, when):
    game = gamegen.game(players, strats)
    sched = ExceptionScheduler(game, count, when)
    sgame = schedgame.schedgame(sched)
    with pytest.raises(SchedulerException):
        await innerloop.inner_loop(sgame, restricted_game_size=5)


@pytest.mark.asyncio
@pytest.mark.parametrize('eq_prob', [x / 10 for x in range(11)])
async def test_innerloop_known_eq(eq_prob):
    game = gamegen.sym_2p2s_known_eq(eq_prob)
    sched = gamesched.gamesched(game)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched), devs_by_role=True)
    assert eqa.size, "didn't find equilibrium"
    expected = [eq_prob, 1 - eq_prob]
    assert np.isclose(eqa, expected, atol=1e-3, rtol=1e-3).all(-1).any()
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
@pytest.mark.parametrize('num', [1, 2])
async def test_innerloop_num_eqa(players, strats, num):
    sgame = gamegen.samplegame(players, strats)
    sched = gamesched.samplegamesched(sgame)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched),
        num_equilibria=num, devs_by_role=True)
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
async def test_backups_used():
    """Test that outerloop uses backups

    Since restricted game size is 1, but the only equilibria has support two,
    this must use backups to find an equilibrium."""
    sgame = gamegen.sym_2p2s_known_eq(0.5)
    sched = gamesched.gamesched(sgame)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched), restricted_game_size=1)
    assert eqa.size, "didn't find equilibrium"
    expected = [0.5, 0.5]
    assert np.isclose(eqa, expected, atol=1e-3, rtol=1e-3).all(-1).any()
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
async def test_initial_restrictions():
    """Test that outerloop uses backups

    Since restricted game size is 1, but the only equilibria has support two,
    this must use backups to find an equilibrium."""
    game = gamegen.sym_2p2s_known_eq(0.5)
    sched = gamesched.gamesched(game)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched), initial_restrictions=[[True, True]],
        restricted_game_size=1)
    assert eqa.size, "didn't find equilibrium"
    expected = [0.5, 0.5]
    assert np.isclose(eqa, expected, atol=1e-3, rtol=1e-3).all(-1).any()
    verify_dist_thresh(eqa)


@pytest.mark.asyncio
async def test_nash_failure():
    """With regret thresh of zero, nash will fail"""
    game = gamegen.sym_2p2s_known_eq(1 / 3)
    sched = gamesched.gamesched(game)
    eqa = await innerloop.inner_loop(
        schedgame.schedgame(sched), regret_thresh=0)
    assert not eqa.size


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', games)
@pytest.mark.parametrize('_', range(5))
async def test_at_least_one(players, strats, _):
    """inner loop should always find one equilibrium with at_least one"""
    game = gamegen.game(players, strats)
    eqa = await innerloop.inner_loop(asyncgame.wrap(game), at_least_one=True)
    assert eqa.size


class SchedulerException(Exception):
    """Exception to be thrown by ExceptionScheduler"""
    pass


class ExceptionScheduler(gamesched.RsGameScheduler):
    """Scheduler that allows triggering exeptions on command"""

    def __init__(self, game, error_after, call_type):
        super().__init__(game)
        self._calls = 0
        self._error_after = error_after
        self._call_type = call_type

    async def sample_payoffs(self, profile):
        self._calls += 1
        if self._error_after <= self._calls and self._call_type == 'pre':
            raise SchedulerException
        pay = await super().sample_payoffs(profile)
        if self._error_after <= self._calls and self._call_type == 'post':
            raise SchedulerException
        return pay



import asyncio

import numpy as np


async def deviation_payoffs(sched, mix, num, *, boots=0, chunk_size=None):
    """Bootstrap deviation payoffs

    Parameters
    ----------
    prof_sched : Scheduler
        The scheduler to sample profiles from.
    mix : ndarray
        The mixture to calculate the regret of.
    num : int
        The number of samples to gather. Must be positive.
    boots : int, optional
        The number of bootstrap samples to take. The accuracy of bootstrap is
        independent of this number, but more will reduce the variance of the
        underlying confidence bounds. The default will compute no bootstrap
        gains.
    chunk_size : int, optional
        An implementation detail specifying how frequently profiles are
        scheduled since this algorithm inherently operates in a streaming
        manner. Ideally this number should be set such that the time to
        schedule and process chunk_size roughly equals the time for one
        simulation. It also controls how much memory this uses. By default this
        is set to ten times the number of bootstraps, or 1000 if no bootstraps
        are requested.

    Notes
    -----
    This uses memory on the order of `boots + chunk_size`. It is inefficient if
    `num` is less than boots.

    Returns
    -------
    mean_gains : ndarray (num_strats,)
        The mean deviation payoffs from the mixture.
    boot_gains : ndarray (boots, num_strats)
        The deviation payoffs for each bootstrap sample.
    """
    assert num > 0, "can't schedule zero samples"
    mix = np.asarray(mix, float)
    chunk_size = chunk_size or boots * 10 or 1000
    devs = np.empty(mix.size)
    mean_devs = np.zeros(mix.size)
    boot_devs = np.zeros((boots, mix.size))
    remaining = np.empty(boots, int)
    remaining.fill(num)

    # XXX This could be made less awkward, but it would help to require python
    # 3.6
    i = 0
    futures = []

    async def update():
        nonlocal i
        fiter = iter(futures)
        for _ in range(len(futures) // sched.num_strats):
            for j in range(sched.num_strats):
                pay = await next(fiter)
                devs[j] = pay[j]
            np.add((devs - mean_devs) / (i + 1), mean_devs, mean_devs)
            samps = np.random.binomial(remaining, 1 / (num - i))
            np.subtract(remaining, samps, remaining)
            np.add(samps[:, None] * devs / num, boot_devs, boot_devs)
            i += 1

    n = num
    while 0 < n:
        new_profs = sched.random_deviation_profiles(
            min(num, chunk_size), mix).reshape((-1, mix.size))
        n -= chunk_size
        new_futures = [asyncio.ensure_future(sched.sample_payoffs(prof))
                       for prof in new_profs]
        await update()
        futures = new_futures
    await update()

    return mean_devs, boot_devs



import argparse
import json
import logging

import numpy as np
from gameanalysis import nash
from gameanalysis import regret
from gameanalysis import restrict

from egta import schedgame
from egta.script import utils


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'brute', help="""Compute equilibria by sampling all profiles""",
        description="""Samples profiles from the entire game, and then runs
        standard equilibrium finding. For games with a large number of players,
        a reduction should be specified. A list of is returned where each
        element has an "equilibrium" and the corresponding "regret" in the full
        game.""")
    parser.add_argument(
        'scheduler', metavar='<sched-spec>', help="""A scheduler specification,
        see below.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=0.1,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    parser.add_argument(
        '--supp-thresh', metavar='<min-prob>', type=float, default=1e-4,
        help="""Minimum probability for a strategy to be considered in support.
        (default: %(default)g)""")
    parser.add_argument(
        '--restrict', '-r', metavar='<restriction-file>',
        type=argparse.FileType('r'), help="""Specify an optional restricted
        game to sample instead of the whole game. Only deviations from the
        restricted strategy set will be scheduled.""")
    parser.add_argument(
        '--one', action='store_true', help="""Guarantee that an equilibrium is
        found. This may take up to exponential time.""")
    parser.add_argument(
        '--min-reg', action='store_true', help="""Return the minimum regret
        profile found, if none were found below regret threshold.""")
    utils.add_reductions(parser)
    utils.add_scheduler_epilog(parser)
    parser.run = run
    return parser


async def run(args):
    sched = await utils.parse_scheduler(args.scheduler)
    red, red_players = utils.parse_reduction(sched, args)

    rest = (np.ones(sched.num_strats, bool) if args.restrict is None
            else sched.restriction_from_json(json.load(args.restrict)))

    async with sched:
        data = await schedgame.schedgame(
            sched, red, red_players).get_deviation_game(rest)

    # now find equilibria
    eqa = sched.trim_mixture_support(
        restrict.translate(nash.mixed_nash(
            data.restrict(rest), regret_thresh=args.regret_thresh,
            dist_thresh=args.dist_thresh, at_least_one=args.one,
            min_reg=args.min_reg), rest),
        thresh=args.supp_thresh)
    reg_info = []
    for eqm in eqa:
        gains = regret.mixture_deviation_gains(data, eqm)
        bri = np.argmax(gains)
        reg_info.append((gains[bri],) + sched.role_strat_names[bri])

    logging.error(
        "brute sampling finished finding %d equilibria:\n%s",
        eqa.shape[0], '\n'.join(
            '{:d}) {} with regret {:g} to {} {}'.format(
                i, sched.mixture_to_repr(eqm), reg, role, strat)
            for i, (eqm, (reg, role, strat))
            in enumerate(zip(eqa, reg_info), 1)))

    json.dump([{'equilibrium': sched.mixture_to_json(eqm),
                'regret': reg,
                'best_response': {'role': role, 'strat': strat}}
               for eqm, (reg, role, strat)
               in zip(eqa, reg_info)], args.output)
    args.output.write('\n')



import numpy as np

from egta import profsched


class CanonScheduler(profsched.Scheduler):
    """A scheduler that removes single strategy roles

    Parameters
    ----------
    sched : Scheduler
        The base scheduler that generates payoffs.
    """

    def __init__(self, sched):
        role_mask = sched.num_role_strats > 1
        super().__init__(
            tuple(r for r, m in zip(sched.role_names, role_mask) if m),
            tuple(s for s, m in zip(sched.strat_names, role_mask) if m),
            sched.num_role_players[role_mask])
        self._sched = sched
        self._players = sched.num_role_players[~role_mask]
        self._inds = np.cumsum(
            role_mask * sched.num_role_strats)[~role_mask]
        self._mask = role_mask.repeat(sched.num_role_strats)

    async def sample_payoffs(self, profile):
        full_prof = np.insert(profile, self._inds, self._players)
        full_pay = await self._sched.sample_payoffs(full_prof)
        return full_pay[self._mask]

    def __str__(self):
        return str(self._sched)


def canon(sched):
    return CanonScheduler(sched)



import asyncio
import pytest

from gameanalysis import gamegen
from gameanalysis import rsgame

from egta import gamesched
from egta import savesched


@pytest.mark.asyncio
@pytest.mark.parametrize('_', range(20))
async def test_basic_profile(_):
    sgame = gamegen.samplegame([4, 3], [3, 4], 0)
    profs = sgame.all_profiles()

    sched = savesched.savesched(gamesched.samplegamesched(sgame))
    assert rsgame.emptygame_copy(sgame) == rsgame.emptygame_copy(sched)
    await asyncio.gather(*[sched.sample_payoffs(p) for p in profs[:10]])
    sched.get_game()

    sched = savesched.savesched(gamesched.samplegamesched(sgame))
    await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    savegame = sched.get_game()
    assert sgame == savegame
    assert sgame == sched.get_game()



import argparse
import json
import logging

import numpy as np

from egta import bootstrap
from egta.script import utils


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'bootstrap', aliases=['boot'], help="""Compute the regret and surplus
        of a mixture""", description="""Samples profiles to compute a sample
        regret and sample surplus of the mixture. By optionally specifying
        percentiles, bootstrap confidence bounds will be returned. The result
        is a json dictionary mapping surplus and regret to either "mean" or a
        string representation of the percentile.""")
    parser.add_argument(
        'scheduler', metavar='<sched-spec>', help="""A scheduler specification,
        see below.""")
    parser.add_argument(
        'mixture', metavar='<mixture-file>', type=argparse.FileType('r'),
        help="""A file with the json formatted mixture to compute the regret
        of.""")
    parser.add_argument(
        'num', metavar='<num-samples>', type=int, help="""The number of samples
        to compute for the regret. One sample for each strategy in the game
        will be taken.""")
    parser.add_argument(
        '--percentile', '-p', metavar='<percentile>', default=[],
        action='append', type=float, help="""A confidence percentile to
        compute.  Specifying at least once switches this to use a bootstrap
        method which takes memory proportional to the number of bootstrap
        samples to run.""")
    parser.add_argument(
        '--boots', '-b', metavar='<num-samples>', type=int, default=101,
        help="""The number of bootstrap samples to take if percentiles is
        specified. More samples will produce less variable confidence bounds at
        the cost of more time. (default: %(default)d)""")
    parser.add_argument(
        '--chunk-size', '-c', metavar='<chunk-size>', type=int, help="""Roughly
        the number of profiles to be scheduled simultaneously. This will be set
        to 10 * boots or 1000 if unspecified. The time to run a simulation
        should roughly correspond to the time to process chunk_size payoffs.
        Keeping this on the order of the number of bootstraps keeps the memory
        requirement constant.""")
    parser.add_argument(
        '--standard', action='store_true', help="""Force output to be
        consistent irrespective of if percentiles is specified or the game is
        symmetric.""")
    utils.add_scheduler_epilog(parser)
    parser.run = run
    return parser


async def run(args):
    sched = await utils.parse_scheduler(args.scheduler)
    if not args.percentile:
        args.boots = 0
    mix = sched.mixture_from_json(json.load(args.mixture))
    async with sched:
        means, boots = await bootstrap.deviation_payoffs(
            sched, mix, args.num, boots=args.boots, chunk_size=args.chunk_size)

    exp_means = np.add.reduceat(means * mix, sched.role_starts)
    exp_boots = np.add.reduceat(boots * mix, sched.role_starts, 1)

    gain_means = means - exp_means.repeat(sched.num_role_strats)
    gain_boots = boots - exp_boots.repeat(sched.num_role_strats, 1)

    role_ind_reg_means = np.fromiter(map(np.argmax, np.split(
        gain_means, sched.role_starts[1:])), int, sched.num_roles)
    role_reg_means = gain_means[role_ind_reg_means + sched.role_starts]
    ind_reg_means = np.argmax(role_reg_means)
    reg_means = role_reg_means[ind_reg_means]

    role_reg_boots = np.maximum.reduceat(gain_boots, sched.role_starts, 1)
    reg_boots = role_reg_boots.max(1)

    role_surp_means = exp_means * sched.num_role_players
    surp_means = role_surp_means.sum()
    role_surp_boots = exp_boots * sched.num_role_players
    surp_boots = role_surp_boots.sum(1)

    reg_percs = np.percentile(reg_boots, args.percentile)
    surp_percs = np.percentile(surp_boots, args.percentile)

    logging.error("bootstrap regret finished with regret %g and surplus %g%s",
                  reg_means, surp_means, '' if not args.percentile else
                  ':\nPerc   Regret    Surplus\n----  --------  --------\n' +
                  '\n'.join('{: 3g}%  {: 8.4g}  {: 8.4g}'.format(p, r, s)
                            for p, r, s
                            in zip(args.percentile, reg_percs, surp_percs)))

    # format output
    if sched.is_symmetric() and not args.standard:
        result = {'surplus': surp_means,
                  'regret': reg_means,
                  'response': sched.strat_names[0][role_ind_reg_means[0]]}

        if args.percentile:
            result = {'mean': result}
            for p, surp, reg in zip(args.percentile, surp_percs, reg_percs):
                result['{:g}'.format(p)] = {'surplus': surp, 'regret': reg}
    else:
        mean_dev = '{}: {}'.format(
            sched.role_names[ind_reg_means],
            sched.strat_names[ind_reg_means][role_ind_reg_means[
                ind_reg_means]])
        result = {'total': {'surplus': surp_means,
                            'regret': reg_means,
                            'response': mean_dev}}
        for role, strats, surp, reg, dev in zip(
                sched.role_names, sched.strat_names, role_surp_means,
                role_reg_means, role_ind_reg_means):
            result[role] = {'surplus': surp,
                            'regret': reg,
                            'response': strats[dev]}

        if args.percentile or args.standard:
            result = {'mean': result}

        for p, surp, reg, role_surps, role_regs in zip(
                args.percentile, surp_percs, reg_percs,
                np.percentile(role_surp_boots, args.percentile, 0),
                np.percentile(role_reg_boots, args.percentile, 0)):
            perc = {'total': {'surplus': surp,
                              'regret': reg}}
            for role, surp, reg in zip(sched.role_names, role_surps,
                                       role_regs):
                perc[role] = {'surplus': surp, 'regret': reg}
            result['{:g}'.format(p)] = perc

    json.dump(result, args.output)
    args.output.write('\n')



import asyncio

import numpy as np

from egta import profsched


class CountScheduler(profsched.Scheduler):
    """A scheduler that wraps each profile in the mean of n profiles

    Parameters
    ----------
    sched : Scheduler
        The base scheduler that generates payoffs.
    count : int > 0
        The number of times a scheduler in the base scheduler should be sampled
        for each payoff returned by this scheduler.
    """

    def __init__(self, sched, count):
        super().__init__(
            sched.role_names, sched.strat_names, sched.num_role_players)
        assert count > 0, "count must be positive {:d}".format(count)
        self._sched = sched
        self._count = count

    async def sample_payoffs(self, profile):
        payoffs = await asyncio.gather(*[
            self._sched.sample_payoffs(profile) for _ in range(self._count)])
        payoff = np.zeros(self.num_strats)
        for i, pay in enumerate(payoffs, 1):
            payoff += (pay - payoff) / i
        return payoff

    def __str__(self):
        return str(self._sched)


def countsched(sched, count):
    return CountScheduler(sched, count)



from egtaonline import api
from gameanalysis import rsgame

from egta import eosched


async def create_scheduler(
        game: "The game id of the game to get profiles from. (required)"=None,
        mem: "Memory in MB for the scheduler. (required)"=None,
        time: "Time in seconds for an observation. (required)"=None,
        auth: "Auth string for egtaonline."=None,
        count: """This scheduler is more efficient at sampling several
        profiles."""='1',
        sleep: """Time to wait in seconds before querying egtaonline to see if
        profiles are complete. Due to the way flux and egtaonline schedule
        jobs, this never needs to be less than 300."""='600',
        max: """The maximum number of jobs to have active on egtaonline at any
        specific time."""='100', **_):
    """A scheduler that gets payoff data from egtaonline."""
    assert game is not None, "`game` must be specified"
    assert mem is not None, "`mem` must be specified"
    assert time is not None, "`time` must be specified"
    game_id = int(game)
    mem = int(mem)
    time = int(time)
    count = int(count)
    sleep = float(sleep)
    max_sims = int(max)

    async with api.api(auth_token=auth) as egta:
        game = await egta.get_game(game_id)
        summ = await game.get_summary()
    game = rsgame.emptygame_json(summ)
    return ApiWrapper(game, egta, game_id, sleep, count, max_sims, mem, time)


class ApiWrapper(eosched.EgtaOnlineScheduler):
    def __init__(self, game, api, *args, **kwargs):
        super().__init__(game, api, *args, **kwargs)
        self.api = api

    async def __aenter__(self):
        await self.api.__aenter__()
        return await super().__aenter__()

    async def __aexit__(self, *args):
        await super().__aexit__(*args)
        await self.api.__aexit__(*args)



import numpy as np
import pytest
from gameanalysis import gamegen

from egta import bootstrap
from egta import gamesched


sizes = [
    [[2], [2]],
    [[2, 2], [2, 2]],
    [[3, 2], [2, 3]],
    [[3, 4], [4, 3]],
]


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
async def test_random_mean_reg(players, strats):
    game = gamegen.game(players, strats)
    mix = game.random_mixture()
    sched = gamesched.gamesched(game)
    mean, boot = await bootstrap.deviation_payoffs(sched, mix, 20)
    assert mean.shape == (game.num_strats,)
    assert boot.shape == (0, game.num_strats)


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
async def test_random_boot_reg(players, strats):
    game = gamegen.game(players, strats)
    mix = game.random_mixture()
    devs = game.deviation_payoffs(mix)
    sched = gamesched.gamesched(game)
    mean, boot = await bootstrap.deviation_payoffs(
        sched, mix, 20, boots=101, chunk_size=5)
    assert mean.shape == (game.num_strats,)
    assert boot.shape == (101, game.num_strats)
    # These aren't guaranteed to be false, but it's incredibly unlikely
    assert not np.allclose(devs, mean)
    assert not np.allclose(devs, boot)


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
async def test_random_pure_boot_reg(players, strats):
    game = gamegen.game(players, strats)
    sched = gamesched.gamesched(game)
    for mix in game.pure_mixtures():
        devs = game.deviation_payoffs(mix)
        mean, boot = await bootstrap.deviation_payoffs(
            sched, mix, 20, boots=101)
        assert np.allclose(devs, mean)
        assert np.allclose(devs, boot)
        assert mean.shape == (game.num_strats,)
        assert boot.shape == (101, game.num_strats)



import asyncio
import functools
import heapq
import logging
from concurrent import futures

import numpy as np
from gameanalysis import collect
from gameanalysis import nash
from gameanalysis import regret
from gameanalysis import restrict


# TODO There's something to be said about individual rationality constraints
# relative to best deviation, i.e. if all are positive gains, but negative
# payoff, that might mean we should warn, or at least explore something else.


# FIXME Should nash_procs just be the executor instead? Although a thread
# executor might make it fail?
async def inner_loop(
        agame, *, initial_restrictions=None, regret_thresh=1e-3,
        dist_thresh=0.1, support_thresh=1e-4, restricted_game_size=3,
        num_equilibria=1, num_backups=1, devs_by_role=False,
        at_least_one=False, nash_procs=2):
    """Inner loop a game using a scheduler

    Parameters
    ----------
    game : RsGame
        The game to find equilibria in. This function is most useful when game
        is a SchedulerGame, but any complete RsGame will work.
    initial_restriction : [[bool]], optional
        Initial restrictions to start inner loop from. If unspecified, every
        pure restriction is used.
    regret_thresh : float > 0, optional
        The maximum regret to consider an equilibrium an equilibrium.
    dist_thresh : float > 0, optional
        The minimum norm between two mixed profiles for them to be considered
        distinct.
    support_thresh : float > 0, optional
        Candidate equilibria strategies with probability lower than this will
        be truncated. This is useful because often Nash finding returns
        strategies with very low support, which now mean extra deviating
        strategies need to be sampled. Trimming saves these samples, but may
        increase regret above the threshold.
    restricted_game_size : int > 0, optional
        The maximum restricted game support size with which beneficial
        deviations must be explored. Restricted games with support larger than
        this are queued and only explored in the event that no equilibrium can
        be found in beneficial deviations smaller than this.
    num_equilibria : int > 0, optional
        The number of equilibria to attempt to find. Only one is guaranteed,
        but this might be beneifical if the game has a known degenerate
        equilibria, but one which is still helpful as a deviating strategy.
    num_backups : int > 0, optional
        In the event that no equilibrium can be found in beneficial deviations
        to small restricted games, other restrictions will be explored. This
        parameter indicates how many restricted games for each role should be
        explored.
    devs_by_role : boolean, optional
        If specified, deviations will only be explored for each role in order,
        proceeding to the next role only when no beneficial deviations are
        found. This can reduce the number of profiles sampled, but may also
        fail to find certain equilibria due to the different path through
        restricted games.
    at_least_one : boolean, optional
        If specified, at least one equilibria will be found in each restricted
        game.  This has the potential to run for a very long time as it may
        take exponential time. If your regret threshold is not set too log for
        your game, this is relatively reasonable though.
    nash_procs : int, optional
        The number of processes to use for finding Nash equilibria. Setting
        None will result in using all of the available processes.
    """
    init_role_dev = 0 if devs_by_role else None

    exp_restrictions = collect.bitset()
    exp_restrictions.add(np.zeros(agame.num_strats, bool))
    backups = [[] for _ in range(agame.num_roles)]
    equilibria = collect.mcces(dist_thresh)
    loop = asyncio.get_event_loop()

    async def add_restriction(rest):
        if not exp_restrictions.add(rest):
            return  # already explored
        if agame.is_pure_restriction(rest):
            # Short circuit for pure restriction
            await add_deviations(rest, rest.astype(float), init_role_dev)
            return
        data = await agame.get_restricted_game(rest)
        reqa = await loop.run_in_executor(executor, functools.partial(
            nash.mixed_nash, data, regret_thresh=regret_thresh,
            dist_thresh=dist_thresh, at_least_one=at_least_one))
        if reqa.size:
            eqa = restrict.translate(data.trim_mixture_support(reqa), rest)
            await asyncio.gather(*[
                add_deviations(rest, eqm, init_role_dev) for eqm in eqa])
        else:
            logging.warning(
                "%s: couldn't find equilibria in restricted game %s. This is "
                "likely due to high variance in payoffs which means "
                "quiesce should be re-run with more samples per profile. "
                "This could also be fixed by performing a more expensive "
                "equilibria search to always return one.",
                agame, agame.restriction_to_repr(rest))

    async def add_deviations(rest, mix, role_index):
        # We need the restriction here, since trimming support may increase
        # regret of strategies in the initial restriction
        data = await agame.get_deviation_game(mix > 0, role_index)
        devs = data.deviation_payoffs(mix)
        exp = np.add.reduceat(devs * mix, agame.role_starts)
        gains = devs - exp.repeat(agame.num_role_strats)
        if role_index is None:
            if np.all((gains <= regret_thresh) | rest):
                # Found equilibrium
                reg = gains.max()
                if equilibria.add(mix, reg):
                    logging.warning(
                        "%s: found equilibrium %s with regret %f", agame,
                        agame.mixture_to_repr(mix), reg)
            else:
                await asyncio.gather(*[
                    queue_restrictions(rgains, ri, rest)
                    for ri, rgains in enumerate(np.split(
                            gains, agame.role_starts[1:]))])

        else:  # Set role index
            rgains = np.split(gains, agame.role_starts[1:])[role_index]
            rrest = np.split(rest, agame.role_starts[1:])[role_index]
            if np.all((rgains <= regret_thresh) | rrest):  # No deviations
                role_index += 1
                if role_index < agame.num_roles:  # Explore next deviation
                    await add_deviations(rest, mix, role_index)
                else:  # found equilibrium
                    # This should not require scheduling as to get here all
                    # deviations have to be scheduled
                    data = await agame.get_deviation_game(mix > 0)
                    reg = regret.mixture_regret(data, mix)
                    if equilibria.add(mix, reg):
                        logging.warning(
                            "%s: found equilibrium %s with regret %f", agame,
                            agame.mixture_to_repr(mix), reg)
            else:
                await queue_restrictions(rgains, role_index, rest)

    async def queue_restrictions(role_gains, role_index, rest):
        role_rest = np.split(rest, agame.role_starts[1:])[role_index]
        if role_rest.all():
            return  # Can't deviate

        rest_size = rest.sum()
        rs = agame.role_starts[role_index]

        br = np.nanargmax(np.where(role_rest, np.nan, role_gains))
        if (role_gains[br] > regret_thresh
                and rest_size < restricted_game_size):
            br_sub = rest.copy()
            br_sub[rs + br] = True
            await add_restriction(br_sub)
        else:
            br = None  # Add best response to backup

        back = backups[role_index]
        for si, (gain, r) in enumerate(zip(role_gains, role_rest)):
            if si == br or r or gain <= 0:
                continue
            sub = rest.copy()
            sub[rs + si] = True
            # XXX Tie id to deterministic random source
            heapq.heappush(back, (-gain, id(sub), sub))  # id for tie-breaking

    restrictions = (
        agame.pure_restrictions() if initial_restrictions is None
        else np.asarray(initial_restrictions, bool))

    with futures.ProcessPoolExecutor(nash_procs) as executor:
        iteration = 0
        while (len(equilibria) < num_equilibria
               and (any(q for q in backups) or
                    not next(iter(exp_restrictions)).all())):
            if iteration == 1:
                logging.warning(
                    "%s: scheduling backup restrictions. This only happens "
                    "when quiesce criteria could not be met with current "
                    "maximum restriction size (%d). This probably means that "
                    "the maximum restriction size should be increased. If "
                    "this is happening frequently, increasing the number of "
                    "backups taken at a time might be desired (currently %s).",
                    agame, restricted_game_size, num_backups)
            elif iteration > 1:
                logging.info("%s: scheduling backup restrictions", agame)

            await asyncio.gather(*[
                add_restriction(r) for r in restrictions])

            restrictions = collect.bitset(exp_restrictions)
            for r, back in enumerate(backups):
                unscheduled = num_backups
                while unscheduled > 0 and back:
                    rest = heapq.heappop(back)[-1]
                    unscheduled -= restrictions.add(rest)
                for _ in range(unscheduled):
                    added = False
                    for mask in restrictions:
                        rmask = np.split(mask, agame.role_starts[1:])[r]
                        if rmask.all():
                            continue
                        rest = mask.copy()
                        # TODO We could randomize instead of taking the first
                        # strategy, but this would remove reproducability
                        s = np.split(rest, agame.role_starts[1:])[r].argmin()
                        rest[agame.role_starts[r] + s] = True
                        restrictions.add(rest)
                        added = True
                        break
                    if not added:
                        break
            iteration += 1

    # Return equilibria
    if equilibria:
        return np.stack([eqm for eqm, _ in equilibria])
    else:
        return np.empty((0, agame.num_strats))  # pragma: no cover



import asyncio
import pytest

import numpy as np
from gameanalysis import gamegen
from gameanalysis import utils

from egta import countsched
from egta import gamesched
from egta import savesched


@pytest.mark.asyncio
async def test_basic_profile():
    sgame = gamegen.samplegame([4, 3], [3, 4])
    profs = utils.unique_axis(sgame.random_profiles(20))

    save = savesched.savesched(gamesched.samplegamesched(sgame))
    sched = countsched.countsched(save, 10)
    assert str(sched) is not None
    paylist = await asyncio.gather(*[
        sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)

    savegame = save.get_game()
    assert list(savegame.num_samples) == [10]



"""A scheduler that gets payoffs from a local simulation"""
import asyncio
import itertools
import json
import logging
import os
import shutil
import tempfile
import zipfile

from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis import utils

from egta import profsched


class ZipScheduler(profsched.Scheduler):
    """Schedule profiles using am EGTA Online zip file

    Parameters
    ----------
    game : RsGame
        A gameanalysis game that indicates how array profiles should be turned
        into json profiles.
    config : {key: value}
        A dictionary mapping string keys to values that will be passed to the
        simulator in the standard simulation spec format.
    zipf : string, file-like
        A zip file that follows the same semantics that EGTA Online expects.
    max_procs : int, optional
        The maximum number of processes to spawn for simulations.
    """

    def __init__(self, game, conf, zipf, *, max_procs=4, simultaneous_obs=1):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._game = paygame.game_copy(rsgame.emptygame_copy(game))
        self.conf = conf
        self.zipf = zipf

        self._extra_profs = {}
        self._base = {}
        self._count = simultaneous_obs
        self._is_open = False
        self._sim_dir = None
        self._prof_dir = None
        self._sim_root = None

        self._num = 0
        self._procs = asyncio.Semaphore(max_procs)

    # FIXME Is it possible to do file ops async
    async def sample_payoffs(self, profile):
        assert self._is_open, "must enter scheduler"
        hprof = utils.hash_array(profile)
        counter, queue = self._extra_profs.get(hprof, (None, None))
        if counter is not None:
            # Already scheduling some profiles
            if next(counter) >= self._count:
                self._extra_profs.pop(hprof)
            pay = await queue.get()
            logging.debug(
                "read payoff for profile: %s", self.profile_to_repr(profile))
            return pay

        else:
            # Need to schedule new profiles
            direc = os.path.join(self._prof_dir.name, str(self._num))
            self._num += 1
            queue = asyncio.Queue()
            if self._count > 1:
                self._extra_profs[hprof] = (itertools.count(2), queue)
            os.makedirs(direc)
            self._base['assignment'] = self._game.profile_to_assignment(
                profile)
            with open(os.path.join(direc, 'simulation_spec.json'),
                      'w') as f:
                json.dump(self._base, f)
            logging.debug(
                "scheduled %d profile%s: %s", self._count,
                '' if self._count == 1 else 's', self.profile_to_repr(profile))

            # Limit simultaneous processes
            async with self._procs:
                proc = await asyncio.create_subprocess_shell(
                    '{} {} {:d}'.format(os.path.join('script', 'batch'),
                                        direc, self._count),
                    cwd=self._sim_root, stderr=asyncio.subprocess.PIPE)
                _, err = await proc.communicate()
            assert proc.returncode == 0, \
                "process failed with returncode {:d} and stderr {}".format(
                    proc.returncode, err)
            obs_files = (
                f for f in os.listdir(direc)
                if 'observation' in f and f.endswith('.json'))
            for _ in range(self._count):
                obs_file = next(obs_files, None)
                assert obs_file is not None, \
                    "simulation didn't write enough observation files"
                with open(os.path.join(direc, obs_file)) as f:
                    pay = self._game.payoff_from_json(json.load(f))
                    pay.setflags(write=False)
                    queue.put_nowait(pay)
            obs_file = next(obs_files, None)
            assert obs_file is None, \
                "simulation wrote too many observation files"
            shutil.rmtree(direc)
            pay = queue.get_nowait()
            logging.debug("read payoff for profile: %s",
                          self.profile_to_repr(profile))
            return pay

    def open(self):
        assert not self._is_open
        try:
            self._num = 0
            self._sim_dir = tempfile.TemporaryDirectory()
            self._prof_dir = tempfile.TemporaryDirectory()
            with zipfile.ZipFile(self.zipf) as zf:
                zf.extractall(self._sim_dir.name)
            sim_files = os.listdir(self._sim_dir.name)
            assert len(sim_files) == 1, "improper zip format"
            self._sim_root = os.path.join(self._sim_dir.name, sim_files[0])
            os.chmod(os.path.join(self._sim_root, 'script', 'batch'), 0o700)

            with open(os.path.join(self._sim_root, 'defaults.json')) as f:
                self._base['configuration'] = json.load(f).get(
                    'configuration', {})
            self._base['configuration'].update(self.conf)

            self._is_open = True
        except Exception as ex:
            self.close()
            raise ex

    def close(self):
        self._is_open = False
        self._sim_dir.cleanup()
        self._prof_dir.cleanup()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def __str__(self):
        return self.zipf


def zipsched(game, conf, zipf, *, max_procs=4, simultaneous_obs=1):
    return ZipScheduler(
        game, conf, zipf, max_procs=max_procs,
        simultaneous_obs=simultaneous_obs)



import asyncio

import numpy as np
import pytest
import timeout_decorator
from gameanalysis import gamegen
from gameanalysis import paygame
from gameanalysis import rsgame

from egta import gamesched
from egta import innerloop
from egta import savesched
from egta import asyncgame
from egta import schedgame
from egta import trace


games = [
    ([1], [2]),
    ([2], [2]),
    ([2, 1], [1, 2]),
    ([1, 2], [2, 1]),
    ([2, 2], [2, 2]),
    ([3, 2], [2, 3]),
    ([1, 1, 1], [2, 2, 2]),
]


def verify_complete_traces(traces):
    """Verify that traces are in order and complete"""
    t = 0.0
    for (first, *_, last), _ in traces:
        assert first <= t
        t = max(t, last)
    assert t == 1.0


# These sometimes take a really long time because of at_least_one and many
# innerloops. If it takes more than a minute, just give up.
@pytest.mark.asyncio
@timeout_decorator.timeout(20)
@pytest.mark.xfail(raises=timeout_decorator.timeout_decorator.TimeoutError)
@pytest.mark.parametrize('players,strats', games)
@pytest.mark.parametrize('_', range(5))
async def test_random_trace_game(players, strats, _):
    agame1 = asyncgame.wrap(gamegen.game(players, strats))
    agame2 = asyncgame.wrap(gamegen.game(players, strats))
    traces = await trace.trace_all_equilibria(
        agame1, agame2, at_least_one=True)
    verify_complete_traces(traces)


# These sometimes take a really long time because of at_least_one and many
# innerloops. If it takes more than a minute, just give up.
@pytest.mark.asyncio
@timeout_decorator.timeout(20)
@pytest.mark.xfail(raises=timeout_decorator.timeout_decorator.TimeoutError)
@pytest.mark.parametrize('players,strats', games)
@pytest.mark.parametrize('_', range(5))
async def test_random_trace_sched(players, strats, _):
    sched1 = gamesched.gamesched(gamegen.game(players, strats))
    sched2 = gamesched.gamesched(gamegen.game(players, strats))
    traces = await trace.trace_all_equilibria(
        schedgame.schedgame(sched1), schedgame.schedgame(sched2),
        at_least_one=True)
    verify_complete_traces(traces)


@pytest.mark.asyncio
async def test_sparse_trace():
    """Test that tracing sparsely samples profiles"""
    base = rsgame.emptygame(4, 3)
    game1 = paygame.game_replace(
        base, base.all_profiles(), (base.all_profiles() > 0) * [1, 0, 0])
    game2 = paygame.game_replace(
        base, base.all_profiles(), (base.all_profiles() > 0) * [-0.5, 1.5, 0])
    save1 = savesched.savesched(gamesched.gamesched(game1))
    save2 = savesched.savesched(gamesched.gamesched(game2))

    sgame1 = schedgame.schedgame(save1)
    sgame2 = schedgame.schedgame(save2)

    await asyncio.gather(
        innerloop.inner_loop(sgame1),
        innerloop.inner_loop(sgame2))

    # Assert that innerloop doesn't scheduler all profiles
    assert save1.get_game().num_profiles == 11
    assert save2.get_game().num_profiles == 11

    ((s1, *_, e1), _), ((s2, *_, e2), _) = await trace.trace_all_equilibria(
        sgame1, sgame2)

    # Assert that trace found the expected equilibria
    assert np.isclose(s1, 0)
    assert np.isclose(e1, 1 / 3, atol=1e-3)
    assert np.isclose(s2, 1 / 3, atol=1e-3)
    assert np.isclose(e2, 1)

    # Assert that trace didn't need many extra profiles
    assert save1.get_game().num_profiles == 12
    assert save2.get_game().num_profiles == 12



import numpy as np
from gameanalysis import rsgame
from gameanalysis import paygame

from egta import profsched


class SaveScheduler(profsched.Scheduler):
    """A scheduler that saves all of the payoff data for output later

    Parameters
    ----------
    game : BaseGame
        The base game of the scheduler.
    sched : Scheduler
        The base scheduler to save payoffs from
    """

    def __init__(self, sched):
        super().__init__(
            sched.role_names, sched.strat_names, sched.num_role_players)
        self._sched = sched
        self._game = paygame.samplegame_copy(rsgame.emptygame_copy(self))
        self._profiles = []
        self._payoffs = []

    async def sample_payoffs(self, profile):
        payoff = await self._sched.sample_payoffs(profile)
        self._profiles.append(profile)
        self._payoffs.append(payoff)
        return payoff

    def get_game(self):
        if self._profiles:
            new_profs = np.concatenate([
                self._game.flat_profiles(),
                np.stack(self._profiles)])
            new_pays = np.concatenate([
                self._game.flat_payoffs(),
                np.stack(self._payoffs)])
            self._profiles.clear()
            self._payoffs.clear()
            self._game = paygame.samplegame_replace_flat(
                self._game, new_profs, new_pays)
        return self._game

    def __str__(self):
        return str(self._sched)


def savesched(sched):
    return SaveScheduler(sched)



import asyncio
import contextlib
import logging

from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis import utils as gu

from egta import profsched
from egta import utils as eu


class EgtaOnlineScheduler(profsched.Scheduler):
    """A profile scheduler that schedules through egta online

    Parameters
    ----------
    game : RsGame
        The gameanalysis basegame representing the game to schedule.
    api : EgtaOnlineApi
        The api object to be uased to query EGTA Online.
    game_id : int
        The id of the egtaonline game to use. It must match the setup of game.
        Use the egtaonline `get_or_create_game` if you need to get a game_id
        from a configuration.
    sleep_time : int
        Time in seconds between queries to egtaonline to determine if profiles
        have finished. This should probably be set roughly equal to the time it
        takes for a simulation to run.
    simultaneous_obs : int
        The number of simultaneous observations to schedule at a time. EGTA
        Online will use this when scheduling.
    max_scheduled : int
        The maximum number of observations to schedule simultaneously. Keeping
        this low helps prevent starving others of flux cycles.
    obs_memory : int
        The amount of memory in MB to allocate for each simulation.
    obs_time : int
        The amount of time in seconds to give each simulation to run. Too low
        and long running simulations will get cancelled giving you biased
        samples, too long and it will take longer to schedule jobs on flux.
    """

    def __init__(self, game, api, game_id, sleep_time, simultaneous_obs,
                 max_scheduled, obs_memory, obs_time):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._api = api
        self._game = paygame.samplegame_copy(rsgame.emptygame_copy(game))
        self._game_id = game_id

        self._sleep_time = sleep_time
        self._obs_memory = obs_memory
        self._obs_time = obs_time
        self._simult_obs = simultaneous_obs

        self._is_open = False
        self._profiles = {}
        self._prof_ids = {}
        self._sched = None
        self._fetcher = None
        self._sched_lock = asyncio.Lock()
        self._scheduled = asyncio.BoundedSemaphore(
            max_scheduled * simultaneous_obs)

    def _check_fetcher(self):
        if self._fetcher.done() and self._fetcher.exception() is not None:
            raise self._fetcher.exception()

    async def sample_payoffs(self, profile):
        assert self._is_open, "not open"
        self._check_fetcher()
        hprof = gu.hash_array(profile)
        data = self._profiles.setdefault(
            hprof, ([0], [0], [0], [None], asyncio.Queue()))
        scheduled, _, claimed, prof_id, pays = data
        claimed[0] += 1
        if scheduled[0] < claimed[0]:
            # TODO make requests async
            scheduled[0] += self._simult_obs
            async with self._sched_lock:
                for _ in range(self._simult_obs):
                    await self._scheduled.acquire()
                pid = prof_id[0]
                if pid is not None:
                    await self._sched.remove_profile(pid)
                assignment = self._game.profile_to_repr(profile)
                prof_id[0] = (await self._sched.add_profile(
                    assignment, scheduled[0]))['id']
                if pid is None:
                    self._prof_ids[prof_id[0]] = data
        pay = await pays.get()
        self._check_fetcher()
        return pay

    async def _fetch(self):
        # TODO Make requests async
        try:
            while True:
                logging.info("query scheduler %d for game %d",
                             self._sched['id'], self._game_id)
                info = await self._sched.get_requirements()
                assert info['active'], "scheduler was deactivated"
                reqs = info['scheduling_requirements']
                for req in reqs:
                    prof_id = req['id']
                    scheduled, received, _, _, pays = self._prof_ids[prof_id]
                    if req['current_count'] <= received[0]:
                        continue
                    egta_prof = await self._api.get_profile(prof_id)
                    jobs = await egta_prof.get_observations()
                    # TODO Is this still necessary
                    # valid = all(o['symmetry_groups'] is not None for o
                    #             in jobs['observations'])
                    obs = self._game.samplepay_from_json(jobs)
                    num = obs.shape[0] - received[0]
                    # Only un-schedule the amount different than what you
                    # thought you scheduled. We can get other observations from
                    # other schedulers.
                    for _ in range(min(num, scheduled[0] - received[0])):
                        self._scheduled.release()
                    received[0] += num
                    obs = obs[:num].copy()
                    obs.setflags(write=False)
                    for o in obs:
                        pays.put_nowait(o)
                await asyncio.sleep(self._sleep_time)
        except Exception as ex:
            for _, (received,), (claimed,), _, pays in self._profiles.values():
                for _ in range(claimed - received):
                    pays.put_nowait(None)
            raise ex

    async def open(self):
        assert not self._is_open, "already open"
        try:
            game = await self._api.get_game(self._game_id)
            obs = await game.get_observations()
            assert (rsgame.emptygame_copy(self._game) ==
                    rsgame.emptygame_json(obs)), \
                "egtaonline game didn't match specified game"
            conf = dict(obs.get('configuration', ()) or ())
            profiles = obs.get('profiles', ()) or ()

            # Parse profiles
            num_profs = len(profiles)
            num_pays = 0
            for jprof in profiles:
                pid = jprof['id']
                prof, spays = self._game.profsamplepay_from_json(jprof)
                spays.setflags(write=False)
                hprof = gu.hash_array(prof)
                pays = asyncio.Queue()
                num_spays = len(spays)
                num_pays += num_spays
                for pay in spays:
                    pays.put_nowait(pay)
                data = ([num_spays], [num_spays], [0], [pid], pays)
                self._profiles[hprof] = data
                self._prof_ids[pid] = data
            logging.info(
                "found %d existing profiles with %d payoffs in game %d",
                num_profs, num_pays, self._game_id)

            # Create and start scheduler
            self._sched = await obs.create_generic_scheduler(
                'egta_' + eu.random_string(20), True, self._obs_memory,
                self._obs_time, self._simult_obs, 1, conf)
            logging.warning(
                "created scheduler %d for running simulations of game %d: "
                "https://%s/generic_schedulers/%d", self._sched['id'],
                self._game_id, self._api.domain, self._sched['id'])
            self._fetcher = asyncio.ensure_future(self._fetch())
            self._is_open = True
        except Exception as ex:
            await self.close()
            raise ex
        return self

    async def close(self):
        if self._fetcher is not None:
            self._fetcher.cancel()
            with contextlib.suppress(Exception):
                await self._fetcher
            self._fetcher = None

        if self._sched is not None:
            await self._sched.deactivate()
            logging.info("deactivated scheduler %d for game %d",
                         self._sched['id'], self._game_id)
            self._sched = None

        if self._sched_lock.locked():
            self._sched_lock.release()
        while True:
            try:
                self._scheduled.release()
            except ValueError:
                break  # Fully reset
        self._profiles.clear()
        self._prof_ids.clear()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, *args):
        await self.close()

    def __str__(self):
        return str(self._game_id)


def eosched(
        game, api, game_id, sleep_time, simultaneous_obs, max_scheduled,
        obs_memory, obs_time):
    return EgtaOnlineScheduler(
        game, api, game_id, sleep_time, simultaneous_obs, max_scheduled,
        obs_memory, obs_time)



import asyncio
import itertools
import logging

from gameanalysis import rsgame
from gameanalysis import trace

from egta import asyncgame
from egta import innerloop


# TODO Expose max step
async def trace_all_equilibria(
        agame0, agame1, regret_thresh=1e-3, **innerloop_args):
    assert rsgame.emptygame_copy(agame0) == rsgame.emptygame_copy(agame1)
    trace_args = dict(regret_thresh=regret_thresh)
    innerloop_args.update(trace_args)

    async def trace_eqm(eqm, t):
        supp = eqm > 0
        game0, game1 = await asyncio.gather(
            agame0.get_deviation_game(supp),
            agame1.get_deviation_game(supp))
        # TODO Should this be wrapped in an executor?
        return trace.trace_equilibria(
            game0, game1, t, eqm, **trace_args)

    async def trace_between(lower, upper):
        if upper <= lower:
            return ()
        mid = (lower + upper) / 2

        midgame = asyncgame.merge(agame0, agame1, mid)
        eqa = await innerloop.inner_loop(midgame, **innerloop_args)

        if not eqa.size:  # pragma: no cover
            logging.warning("%s: found no equilibria for t: %g", midgame, mid)
            return ()

        # XXX This shouldn't really be async as all data should be there, could
        # potentially add a get_nowait to async game
        traces = await asyncio.gather(*[
            trace_eqm(eqm, mid) for eqm in eqa])

        lupper = min(t[0] for t, _ in traces)
        ulower = max(t[-1] for t, _ in traces)
        logging.warning(
            "%s: traced %g out to %g - %g", midgame, mid, lupper, ulower)

        lower_traces, upper_traces = await asyncio.gather(
            trace_between(lower, lupper), trace_between(ulower, upper))
        # Lazily extend them
        return itertools.chain(lower_traces, traces, upper_traces)

    traces = await trace_between(0.0, 1.0)
    return sorted(traces, key=lambda tr: (tr[0][0], tr[0][-1]))



import numpy as np
import pytest
from gameanalysis import gamegen
from gameanalysis import restrict
from gameanalysis import rsgame

from egta import asyncgame


@pytest.mark.asyncio
async def test_basic_asyncgame():
    game = gamegen.game([4, 3], [3, 4])
    agame = asyncgame.wrap(game)
    rest = agame.random_restriction()
    rgame = await agame.get_restricted_game(rest)
    assert rgame.is_complete()
    assert (rsgame.emptygame_copy(rgame) ==
            rsgame.emptygame_copy(game.restrict(rest)))

    dgame = await agame.get_deviation_game(rest)
    mix = restrict.translate(rgame.random_mixture(), rest)
    assert not np.isnan(dgame.deviation_payoffs(mix)).any()

    dup = asyncgame.wrap(game)
    assert hash(dup) == hash(agame)
    assert dup == agame


@pytest.mark.asyncio
async def test_merge_asyncgame():
    game0 = gamegen.game([4, 3], [3, 4])
    game1 = gamegen.game([4, 3], [3, 4])
    agame = asyncgame.merge(asyncgame.wrap(game0), asyncgame.wrap(game1), 0.4)

    rest = agame.random_restriction()
    rgame = await agame.get_restricted_game(rest)
    assert rgame.is_complete()
    assert (rsgame.emptygame_copy(rgame) ==
            rsgame.emptygame_copy(game0.restrict(rest)))

    dgame = await agame.get_deviation_game(rest)
    mix = restrict.translate(rgame.random_mixture(), rest)
    assert not np.isnan(dgame.deviation_payoffs(mix)).any()

    dup = asyncgame.merge(asyncgame.wrap(game0), asyncgame.wrap(game1), 0.4)
    assert hash(dup) == hash(agame)
    assert dup == agame



"""Script utility for running equilibrium trace"""
import asyncio
import json
import logging

from gameanalysis import mergegame
from gameanalysis import regret

from egta import canonsched
from egta import schedgame
from egta import trace
from egta.script import utils


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'trace', help="""Compute trace of equilibria between two games""",
        description="""Computes traces of equilibria as the probability of
        mixing between two games changes. This uses quiesce as a subroutine and
        sparsely schedules profiles so that this can be done without having to
        fully sample either game. The results is a list of traces, where each
        trace is a list of the mixture probability, the equilibriium, and the
        regret. In order to allow this to work on games with singleton players,
        games are "normalized" by removing them prior to scheduling. Thus, any
        roles with only one strategy should be omitted from reductions, and
        will also be omitted from equilibria as their answer is trivial.""")
    parser.add_argument(
        'sched0', metavar='<sched-spec-0>', help="""The scheduler specification
        for the game when t is 0. See below.""")
    parser.add_argument(
        'sched1', metavar='<sched-spec-1>', help="""The scheduler specification
        for the game when t is 1. See below.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        This can't be strictly enforced as the ODE that this solves has some
        numerical instability, and there is no way to guarantee the equilibria
        stays small.  (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=0.1,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    parser.add_argument(
        '--max-restrict-size', metavar='<support>', type=int, default=3,
        help="""Support size threshold, beyond which restricted games are not
        required to be explored.  (default: %(default)d)""")
    parser.add_argument(
        '--num-equilibria', metavar='<num>', type=int, default=1,
        help="""Number of equilibria requested to be found. This is mainly
        useful when game contains known degenerate equilibria, but those
        strategies are still useful as deviating strategies. (default:
        %(default)d)""")
    parser.add_argument(
        '--num-backups', metavar='<num>', type=int, default=1, help="""Number
        of backup restricted strategy set to pop at a time, when no equilibria
        are confirmed in initial required set.  When games get to this point
        they can quiesce slowly because this by default pops one at a time.
        Increasing this number can get games like tis to quiesce more quickly,
        but naturally, also schedules more, potentially unnecessary,
        simulations. (default: %(default)d)""")
    parser.add_argument(
        '--dev-by-role', action='store_true', help="""Explore deviations in
        role order instead of all at once. By default, when checking for
        beneficial deviations, all role deviations are scheduled at the same
        time. Setting this will check one role at a time. If a beneficial
        deviation is found, then that restricted strategy set is scheduled
        without exploring deviations from the other roles.""")
    parser.add_argument(
        '--one', action='store_true', help="""Guarantee that an equilibrium is
        found in every restricted game. This may take up to exponential
        time.""")
    utils.add_reductions(parser)
    utils.add_scheduler_epilog(parser)
    parser.run = run
    return parser


async def run(args):
    sched0 = CanonWrapper(await utils.parse_scheduler(args.sched0))
    sched1 = CanonWrapper(await utils.parse_scheduler(args.sched1))
    red, red_players = utils.parse_reduction(sched0, args)
    agame0 = schedgame.schedgame(sched0, red, red_players)
    agame1 = schedgame.schedgame(sched1, red, red_players)

    async def get_point(t, eqm):
        supp = eqm > 0
        game0 = await agame0.get_deviation_game(supp)
        game1 = await agame1.get_deviation_game(supp)
        reg = regret.mixture_regret(
            mergegame.merge(game0, game1, t), eqm)
        return {
            't': float(t),
            'equilibrium': sched0.mixture_to_json(eqm),
            'regret': float(reg)}

    async def get_trace(ts, teqa):
        return await asyncio.gather(*[
            get_point(t, eqm) for t, eqm in zip(ts, teqa)])

    async with sched0, sched1:
        traces = await trace.trace_all_equilibria(
            agame0, agame1, regret_thresh=args.regret_thresh,
            dist_thresh=args.dist_thresh,
            restricted_game_size=args.max_restrict_size,
            num_equilibria=args.num_equilibria, num_backups=args.num_backups,
            devs_by_role=args.dev_by_role, at_least_one=args.one)
        jtraces = await asyncio.gather(*[
            get_trace(ts, teqa) for ts, teqa in traces])

    max_reg = 0.0
    spans = [[-1, -1]]
    for jtrace in jtraces:
        max_reg = max(max_reg, max(p['regret'] for p in jtrace))
        span = spans[-1]
        start, *_, end = jtrace
        if start['t'] <= span[1]:
            span[1] = max(span[1], end['t'])
        else:
            spans.append([start['t'], end['t']])

    logging.error(
        "tracing finished finding %d traces covering %s, with maximum regret "
        "%g", len(jtraces),
        ' U '.join('[{:g}, {:g}]'.format(s, e) for s, e in spans[1:]), max_reg)

    json.dump(jtraces, args.output)
    args.output.write('\n')


class CanonWrapper(canonsched.CanonScheduler):
    async def __aenter__(self):
        await self._sched.__aenter__()
        return self

    async def __aexit__(self, *args):
        await self._sched.__aexit__(*args)



import numpy as np
import pytest
from gameanalysis import gamegen

from egta import canonsched
from egta import gamesched


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', [
    ([2], [2]), ([1, 2], [2, 1]), ([2, 1], [1, 2]), ([2, 1, 2], [1, 2, 1]),
    ([1, 2, 1], [2, 1, 2]),
])
async def test_basic_profile(players, strats):
    game = gamegen.game(players, strats)
    sched = canonsched.canon(gamesched.gamesched(game))
    assert np.all(sched.num_role_strats > 1)
    pay = await sched.sample_payoffs(sched.random_profile())
    assert pay.size == sched.num_strats



import asyncio
import logging

import numpy as np
from gameanalysis import paygame
from gameanalysis import restrict
from gameanalysis import rsgame
from gameanalysis import utils
from gameanalysis.reduction import identity as idr

from egta import asyncgame


class ReductionSchedulerGame(asyncgame.AsyncGame):
    def __init__(self, sched, red, red_players):
        super().__init__(sched.role_names, sched.strat_names,
                         sched.num_role_players)
        self._sched = sched
        self._rgame = rsgame.emptygame_copy(
            red.reduce_game(rsgame.emptygame_copy(self), red_players))
        self._red = red
        self._profiles = {}

    async def _get_game(self, profs):
        futures = []
        for prof in profs:
            hprof = utils.hash_array(prof)
            future = self._profiles.get(hprof, None)
            if future is None:
                future = asyncio.ensure_future(
                    self._sched.sample_payoffs(prof))
                self._profiles[hprof] = future
            futures.append(future)
        lpays = await asyncio.gather(*futures)
        pays = np.stack(lpays)
        return paygame.game_replace(self, profs, pays)

    def _rprofs(self, rest):
        return restrict.translate(
            self._red.expand_profiles(
                rsgame.emptygame_copy(self).restrict(rest),
                self._rgame.restrict(rest).all_profiles()),
            rest)

    async def get_restricted_game(self, rest):
        logging.info(
            "%s: scheduling restriction %s", self,
            self.restriction_to_repr(rest))
        game = await self._get_game(self._rprofs(rest))
        return _ReductionGame(
            game.restrict(rest), self._red, self._rgame.num_role_players)

    async def get_deviation_game(self, rest, role_index=None):
        logging.info(
            "%s: scheduling deviations from %s%s", self,
            self.restriction_to_repr(rest),
            '' if role_index is None else ' by role ' +
            self.role_names[role_index])
        dprofs = self._red.expand_deviation_profiles(
            self, rest, self._rgame.num_role_players, role_index)
        rprofs = self._rprofs(rest)
        game = await self._get_game(np.concatenate([rprofs, dprofs]))
        return _ReductionGame(game, self._red, self._rgame.num_role_players)

    def __str__(self):
        return str(self._sched)


def schedgame(sched, red=idr, red_players=None):
    return ReductionSchedulerGame(sched, red, red_players)


# TODO This should be moved to game analysis, when the appropriate contract can
# be decided upon. There are a few issues with how we treat the existence of
# profiles with whether or not deviation payoffs exists. For example, maximum
# restrictions checks that profiles are in the game, instead of whether
# deviation data exists... Which is accurate, but not what we would want for
# these style of reduced games.
class _ReductionGame(rsgame.RsGame):
    """A game with only reduced profiles

    This is a wrapper so reduced games look full."""
    def __init__(self, game, reduction, red_players):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._game = game
        self._rgame = reduction.reduce_game(game, red_players)
        self._red = reduction

    @property
    def num_complete_profiles(self):
        return self._game.num_complete_profiles

    @property
    def num_profiles(self):
        return self._game.num_profiles

    def profiles(self):
        return self._game.profiles()

    def payoffs(self):
        return self._game.payoffs()

    def deviation_payoffs(self, mix, *, jacobian=False, **kwargs):
        return self._rgame.deviation_payoffs(
            mix, jacobian=jacobian, **kwargs)

    def get_payoffs(self, profiles):
        self._game.get_payoffs(profiles)

    def max_strat_payoffs(self):
        return self._game.max_strat_payoffs()

    def min_strat_payoffs(self):
        return self._game.min_strat_payoffs()

    # FIXME These could be made more efficient by just taking the reduced game
    # at input and normalizing it here, however, this is a preliminary
    # implementation, and so not worth the efficiency.
    def normalize(self):
        return _ReductionGame(
            self._game.normalize(), self._red, self._rgame.num_role_players)

    def restrict(self, rest):
        return _ReductionGame(
            self._game.restrict(rest), self._red, self._rgame.num_role_players)

    def __contains__(self, profile):
        return profile in self._game

    @utils.memoize
    def __hash__(self):
        return hash((super().__hash__(), self._game, self._rgame))

    def __eq__(self, other):
        return (super().__eq__(other) and
                self._game == other._game and
                self._rgame == other._rgame)



import asyncio
import io
import json
import zipfile

import numpy as np
import pytest
from gameanalysis import rsgame

from egta import countsched
from egta import zipsched


def batch_to_zip(bash):
    buff = io.BytesIO()
    with zipfile.ZipFile(buff, 'w') as zf:
        zf.writestr('package/defaults.json', json.dumps({}))
        zf.writestr('package/script/batch', '#!/usr/bin/env bash\n' + bash)
    return buff


@pytest.fixture
def jgame():
    with open('cdasim/game.json') as f:
        return json.load(f)


@pytest.fixture
def conf(jgame):
    return jgame['configuration']


@pytest.fixture
def game(jgame):
    return rsgame.emptygame_json(jgame)


@pytest.mark.asyncio
async def test_basic_profile(conf, game):
    profs = game.random_profiles(20)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(game, conf, zipf) as sched:
        assert str(sched) is not None
        assert rsgame.emptygame_copy(sched) == game
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)


@pytest.mark.asyncio
async def test_no_conf(game):
    profs = game.random_profiles(20)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(game, {}, zipf) as sched:
        assert rsgame.emptygame_copy(sched) == game
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)


@pytest.mark.asyncio
@pytest.mark.parametrize('count', [2, 3])
async def test_simultaneous_obs(conf, game, count):
    profs = game.random_profiles(10).repeat(2, 0)
    zipf = 'cdasim/cdasim.zip'

    with zipsched.zipsched(
            game, conf, zipf, simultaneous_obs=count) as sched:
        sched = countsched.CountScheduler(sched, count)
        awaited = await asyncio.gather(*[
            sched.sample_payoffs(p) for p in profs])

    pays = np.stack(awaited)
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)


@pytest.mark.asyncio
async def test_get_fail(conf, game):
    prof = game.random_profile()
    zipf = batch_to_zip('sleep 0.2 && false')

    with zipsched.zipsched(game, conf, zipf) as sched:
        future = sched.sample_payoffs(prof)
        with pytest.raises(AssertionError):
            await future


@pytest.mark.asyncio
async def test_bad_zipfile(conf, game):
    zipf = 'nonexistent'
    with pytest.raises(FileNotFoundError):
        with zipsched.zipsched(game, conf, zipf):
            pass  # pragma: no cover


@pytest.mark.asyncio
async def test_no_obs(conf, game):
    prof = game.random_profile()
    zipf = batch_to_zip('sleep 0.2')

    with zipsched.zipsched(game, conf, zipf) as sched:
        future = sched.sample_payoffs(prof)
        with pytest.raises(AssertionError):
            await future



"""python file with a bunch of utility methods"""
import random
import string


def random_string(length, choices=string.ascii_letters + string.digits):
    """Returns a random string of length length.

    Can optinally specify the characters it's drawn from."""
    return ''.join(random.choice(choices) for _ in range(length))



import io
import itertools
import json
import random
import subprocess
import sys
import traceback
from os import path
from unittest import mock

import numpy as np
import pytest
from egtaonline import api
from egtaonline import mockserver
from gameanalysis import gamegen
from gameanalysis import gamereader
from gameanalysis import rsgame
from gameanalysis import utils

from egta import __main__ as main


base = path.dirname(path.realpath(__file__))
DIR = base
EGTA = path.join(DIR, '..', 'bin', 'egta')
SIM_DIR = path.join(DIR, '..', 'cdasim')


async def run(*args):
    """Run a command line and return if it ran successfully"""
    try:
        await main.amain(*args)
    except SystemExit as ex:
        return not int(str(ex))
    except Exception:
        traceback.print_exc()
        return False
    return True


def stdin(inp):
    """Patch stdin with input"""
    return mock.patch.object(sys, 'stdin', io.StringIO(inp))


def stdout():
    """Patch stdout and return stringio"""
    return mock.patch.object(sys, 'stdout', io.StringIO())


def stderr():
    """Patch stderr and return stringio"""
    return mock.patch.object(sys, 'stderr', io.StringIO())


@pytest.fixture(scope='session')
def game_info(tmpdir_factory):
    game_file = str(tmpdir_factory.mktemp('games').join('game.json'))
    game = gamegen.game([3, 2], [2, 3])
    with open(game_file, 'w') as f:
        json.dump(game.to_json(), f)
    return game, 'game:game:{}'.format(game_file)


@pytest.fixture(scope='session')
def sim_game_info(tmpdir_factory):
    sim_dir = path.join(base, '..', 'cdasim')
    game_file = path.join(sim_dir, 'small_game.json')
    sim = [path.join(base, '..', 'bin', 'python'),
           path.join(sim_dir, 'sim.py'), '1', '--single']
    zip_file = path.join(sim_dir, 'cdasim.zip')
    with open(game_file) as f:
        jgame = json.load(f)
    conf_file = str(tmpdir_factory.mktemp('conf').join('conf.json'))
    with open(conf_file, 'w') as f:
        json.dump(jgame['configuration'], f)
    game = rsgame.emptygame_copy(gamereader.loadj(jgame))
    simsched = 'sim:game:{},conf:{},command:{}'.format(
        game_file, conf_file, ' '.join(sim))
    zipsched = 'zip:game:{},zipf:{}'.format(game_file, zip_file)
    return game, simsched, zipsched


def assert_term(sleep1, sleep2, *args):
    proc = subprocess.Popen((EGTA,) + args)
    try:
        proc.wait(sleep1)
        # proc.wait should throw exception and pass i.e. process is still
        # running after sleep1
        proc.kill()  # pragma: no cover
        assert False, "process finished during first sleep"  # pragma: no cover
    except subprocess.TimeoutExpired:
        pass
    proc.terminate()
    try:
        assert proc.wait(sleep2), "process succeeded after termination"
    except subprocess.TimeoutExpired:  # pragma: no cover
        # proc.terminate should terminate process before sleep2 is reached, so
        # we should never trigger except
        proc.kill()
        assert False, "process didn't terminate in second sleep"


@pytest.mark.asyncio
async def test_help():
    assert not await run()
    assert not await run('--fail')
    with stderr() as err:
        assert await run('--help'), err.getvalue()
    with stderr() as err:
        assert await run('brute', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('quiesce', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('boot', '--help'), err.getvalue()
    with stderr() as err:
        assert await run('trace', '--help'), err.getvalue()


@pytest.mark.asyncio
@pytest.mark.parametrize('red', [
    [],
    ['--dpr', 'r0:2,r1:2'],
    ['--hr', 'r0:2,r1:2'],
])
async def test_brute_game(game_info, red):
    game, sched = game_info
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',count:2', *red), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_brute_game_tag(game_info):
    game, sched = game_info
    with stdout() as out, stderr() as err:
        assert await run(
            '--tag', 'test', 'brute', sched + ',count:2'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_brute_game_restriction(game_info):
    game, sched = game_info
    rest = json.dumps(game.restriction_to_json(game.random_restriction()))
    with stdin(rest), stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',count:2', '--restrict', '-'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_brute_game_term(game_info):
    _, sched = game_info
    assert_term(0.5, 0.5, 'brute', sched + ',count:2')


@pytest.mark.asyncio
async def test_prof_data(game_info, tmpdir):
    game, sched = game_info
    prof_file = str(tmpdir.join('profs.json'))
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',sample:,save:' + prof_file), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])
    with open(prof_file) as f:
        prof_game = gamereader.load(f)
    assert rsgame.emptygame_copy(game) == rsgame.emptygame_copy(prof_game)


@pytest.mark.asyncio
async def test_sim(sim_game_info):
    game, sched, _ = sim_game_info
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sched, '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_sim_in(sim_game_info):
    game, sched, _ = sim_game_info
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',game:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip(sim_game_info):
    game, _, sched = sim_game_info
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sched, '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_in(sim_game_info, tmpdir):
    game, _, sched = sim_game_info
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',game:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_conf_in(sim_game_info):
    game, _, sched = sim_game_info
    conf = json.dumps({'max_value': 2})
    with stdin(conf), stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',conf:-', '--dpr', 'buyers:2;sellers:2',
            '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_zip_conf_file(sim_game_info, tmpdir):
    game, _, sched = sim_game_info
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as f:
        json.dump({'max_value': 2}, f)
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', sched + ',conf:' + conf_file, '--dpr',
            'buyers:2;sellers:2', '--min-reg'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_brute_sim_term(sim_game_info):
    _, sched, _ = sim_game_info
    assert_term(0.5, 0.5, 'brute', sched)


@pytest.mark.asyncio
async def test_sim_delayed_fail(tmpdir):
    game = rsgame.emptygame(4, 3)
    game_file = str(tmpdir.join('game.json'))
    with open(game_file, 'w') as f:
        json.dump(game.to_json(), f)
    script = str(tmpdir.join('script.sh'))
    with open(script, 'w') as f:
        f.write('sleep 1 && false')
    sched = 'sim:game:{},command:bash {}'.format(game_file, script)
    with stdin(json.dumps({})):
        assert not await run(
            'brute', sched)


@pytest.mark.asyncio
async def test_sim_conf(sim_game_info, tmpdir):
    game, sched, _ = sim_game_info
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as f:
        json.dump({
            'markup': 'standard',
            'max_value': 1,
            'arrivals': 'simple',
            'market': 'call',
        }, f)
    with stdout() as out, stderr() as err:
        assert await run(
            'brute', '--dpr', 'buyers:2;sellers:2', sched + ',conf:' +
            conf_file), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_sim_conf_in(sim_game_info, tmpdir):
    game, sched, _ = sim_game_info
    conf = json.dumps({
        'markup': 'standard',
        'max_value': 1,
        'arrivals': 'simple',
        'market': 'call',
    })
    with stdin(conf), stdout() as out, stderr() as err:
        assert await run(
            'brute', '--dpr', 'buyers:2;sellers:2',
            sched + ',conf:-'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_innerloop(game_info):
    game, sched = game_info
    with stdout() as out, stderr() as err:
        assert await run('quiesce', sched), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


def test_innerloop_game_term(game_info):
    _, sched = game_info
    assert_term(0.5, 0.5, 'quiesce', sched)


def test_innerloop_sim_term(sim_game_info):
    _, sched, _ = sim_game_info
    assert_term(0.5, 0.5, 'quiesce', sched)


@pytest.mark.asyncio
async def test_innerloop_dpr(game_info):
    game, sched = game_info
    with stdout() as out, stderr() as err:
        assert await run(
            'quiesce', sched, '--dpr', 'r0:2;r1:2'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


@pytest.mark.asyncio
async def test_innerloop_hr(game_info):
    game, sched = game_info
    with stdout() as out, stderr() as err:
        assert await run(
            'quiesce', sched, '--hr', 'r0:2;r1:2'), err.getvalue()
        for eqm in json.loads(out.getvalue()):
            game.mixture_from_json(eqm['equilibrium'])


async def get_eosched(server, game, name):
    async with api.api() as egta:
        sim = await egta.get_simulator(server.create_simulator(
            name, '1', delay_dist=lambda: random.random() / 100))
        await sim.add_strategies(dict(zip(game.role_names, game.strat_names)))
        eogame = await sim.create_game(name, game.num_players)
        await eogame.add_symgroups(list(zip(
            game.role_names, game.num_role_players, game.strat_names)))
        return 'eo:game:{:d},mem:2048,time:60,sleep:0.1'.format(
            eogame['id'])


@pytest.mark.asyncio
async def test_brute_egta_game(game_info):
    game, game_file = game_info
    async with mockserver.server() as server:
        sched = await get_eosched(server, game, 'game')
        with stdout() as out, stderr() as err:
            assert await run(
                'brute', sched, '--dpr', 'r0:2;r1:2',
                '--min-reg'), err.getvalue()
            for eqm in json.loads(out.getvalue()):
                game.mixture_from_json(eqm['equilibrium'])


def verify_trace_json(game, traces):
    for trace in traces:
        for point in trace:
            assert point.keys() == {'regret', 'equilibrium', 't'}
            game.mixture_from_json(point['equilibrium'])
        assert utils.is_sorted(trace, key=lambda t: t['t'])


@pytest.mark.asyncio
@pytest.mark.parametrize('red', [
    [],
    ['--dpr', 'buyers:2,sellers:2'],
    ['--hr', 'buyers:2,sellers:2'],
])
async def test_trace(sim_game_info, tmpdir, red):
    game, sched, _ = sim_game_info
    conf_file = str(tmpdir.join('conf.json'))
    with open(conf_file, 'w') as f:
        json.dump({
            'markup': 'standard',
            'max_value': 1,
            'arrivals': 'simple',
            'market': 'call',
        }, f)
    prof_data = str(tmpdir.join('data.json'))
    with stdout() as out, stderr() as err:
        assert await run(
            'trace', sched + ',save:' + prof_data, sched + ',conf:' +
            conf_file, *red), err.getvalue()
    traces = json.loads(out.getvalue())
    verify_trace_json(game, traces)
    with open(prof_data) as f:
        data = gamereader.load(f)
    assert game == rsgame.emptygame_copy(data)


def add_singleton_role(game, index, role_name, strat_name, num_players):
    """Add a singleton role to a game"""
    role_names = list(game.role_names)
    role_names.insert(index, role_name)
    strat_names = list(game.strat_names)
    strat_names.insert(index, [strat_name])
    role_players = np.insert(game.num_role_players, index, num_players)
    return rsgame.emptygame_names(role_names, role_players, strat_names)


@pytest.mark.asyncio
async def test_trace_norm():
    base_game = rsgame.emptygame([3, 2], [2, 3])
    async with mockserver.server() as server:
        game0 = add_singleton_role(base_game, 0, 'a', 'sx', 1)
        sched0 = await get_eosched(server, game0, 'game0')
        game1 = add_singleton_role(base_game, 1, 'r00', 's0', 3)
        sched1 = await get_eosched(server, game1, 'game1')
        with stdout() as out, stderr() as err:
            assert await run('trace', sched0, sched1), err.getvalue()
        traces = json.loads(out.getvalue())
    verify_trace_json(base_game, traces)


@pytest.mark.asyncio
async def test_boot_game(game_info, tmpdir):
    game, sched = game_info
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as f:
        json.dump(game.mixture_to_json(game.random_mixture()), f)
    with stdout() as out, stderr() as err:
        assert await run(
            'boot', sched, mix_file, '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'total', 'r0', 'r1'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()


@pytest.mark.asyncio
async def test_boot_symmetric(tmpdir):
    game = gamegen.game(4, 3)
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as f:
        json.dump(game.mixture_to_json(game.random_mixture()), f)
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'boot', 'game', mix_file, '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'surplus', 'regret', 'response'} == results.keys()


@pytest.mark.asyncio
async def test_boot_symmetric_percs(tmpdir):
    game = gamegen.game(4, 3)
    mix_file = str(tmpdir.join('mix.json'))
    with open(mix_file, 'w') as f:
        json.dump(game.mixture_to_json(game.random_mixture()), f)
    with stdin(json.dumps(game.to_json())), stdout() as out, stderr() as err:
        assert await run(
            'boot', 'game', mix_file, '10', '-p95'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'95', 'mean'} == results.keys()
    assert {'surplus', 'regret', 'response'} == results['mean'].keys()
    assert {'surplus', 'regret'} == results['95'].keys()


@pytest.mark.asyncio
async def test_boot_game_percs(game_info):
    game, sched = game_info
    mix = json.dumps(game.mixture_to_json(game.random_mixture()))
    with stdin(mix), stdout() as out, stderr() as err:
        assert await run(
            'boot', sched, '-', '20', '--percentile', '95',
            '-p99'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'mean', '99', '95'} == results.keys()
    for val in results.values():
        assert {'total', 'r0', 'r1'} == val.keys()
    for val in results['mean'].values():
        assert {'surplus', 'regret', 'response'} == val.keys()
    for val in itertools.chain(results['95'].values(), results['99'].values()):
        assert {'surplus', 'regret'} == val.keys()


@pytest.mark.asyncio
async def test_boot_sim(sim_game_info):
    game, sched, _ = sim_game_info
    mix = json.dumps(game.mixture_to_json(game.random_mixture()))
    with stdin(mix), stdout() as out, stderr() as err:
        assert await run(
            'boot', sched, '-', '50', '--chunk-size', '10'), err.getvalue()
        results = json.loads(out.getvalue())
    assert {'total', 'buyers', 'sellers'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()



"""Python script for performing egta"""
import asyncio
import argparse
import logging
import smtplib
import sys
import traceback

from egta.script import bootstrap
from egta.script import brute
from egta.script import innerloop
from egta.script import trace


# TODO Create a scheduler that runs jobs on flux, but without going through
# egta online, potentially using spark

async def amain(*argv):
    parser = argparse.ArgumentParser(
        description="""Command line egta. To run, both an equilibrium finding
        method and a profile scheduler must be specified. Each element
        processes the arguments after it, e.g. `egta -a brute -b game -c` will
        pass -a to the base parser, -b to the brute equilibrium solver, and -c
        to the game profile scheduler.""")

    # Standard arguments
    parser.add_argument(
        '--output', '-o', metavar='<output-file>', type=argparse.FileType('w'),
        default=sys.stdout, help="""The file to write the output to. (default:
        stdout)""")
    parser.add_argument(
        '--verbose', '-v', action='count', default=0, help="""Increases the
        verbosity level for standard error.""")
    parser.add_argument(
        '-e', '--email_verbosity', action='count', default=0, help="""Increases
        the verbosity level for emails.""")
    parser.add_argument(
        '-r', '--recipient', metavar='<email-address>', action='append',
        default=[], help="""Specify an email address to receive email logs at.
        Can specify multiple email addresses.""")
    parser.add_argument(
        '--tag', metavar='<tag>', help="""Specify an optional tag that will get
        appended to logs and appear in the email subject.""")

    # All of the actual methods to run
    eq_methods = parser.add_subparsers(
        title='operations', dest='method', metavar='<operation>', help="""The
        operation to run on the game. Available commands are:""")
    for module in [brute, bootstrap, innerloop, trace]:
        module.add_parser(eq_methods)

    # Parse args and process
    args = parser.parse_args(argv)
    method = eq_methods.choices[args.method]

    tag = '' if args.tag is None else args.tag + ' '

    stderr_handle = logging.StreamHandler(sys.stderr)
    stderr_handle.setLevel(50 - 10 * min(args.verbose, 4))
    stderr_handle.setFormatter(logging.Formatter(
        '%(asctime)s {}{} %(message)s'.format(
            args.method, tag)))
    log_handlers = [stderr_handle]

    # Email Logging
    if args.recipient:  # pragma: no cover
        smtp_host = 'localhost'

        # We need to do this to match the from address to the local host name
        # otherwise, email logging will not work. This seems to vary somewhat
        # by machine
        with smtplib.SMTP(smtp_host) as server:
            smtp_fromaddr = 'EGTA Online <egta_online@{host}>'.format(
                host=server.local_hostname)

        email_handler = logging.handlers.SMTPHandler(
            smtp_host, smtp_fromaddr, args.recipient,
            'EGTA Status for {}{}'.format(args.method, tag))
        email_handler.setLevel(50 - args.email_verbosity * 10)
        email_handler.setFormatter(logging.Formatter(
            '%(message)s'))
        log_handlers.append(email_handler)

    logging.basicConfig(level=0, handlers=log_handlers)

    try:
        await method.run(args)

    except KeyboardInterrupt as ex:  # pragma: no cover
        logging.critical('execution interrupted by user')
        raise ex

    except Exception as ex:  # pragma: no cover
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logging.critical(''.join(traceback.format_exception(
            exc_type, exc_value, exc_traceback)))
        raise ex


def main():  # pragma: no cover
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(amain(*sys.argv[1:]))
    finally:
        loop.close()



import argparse
import inspect
import itertools
import json
import shutil
import textwrap

from gameanalysis import reduction

from egta import countsched
from egta import savesched
from egta.script import eosched
from egta.script import gamesched
from egta.script import simsched
from egta.script import zipsched


def add_reductions(parser):
    reductions = parser.add_mutually_exclusive_group()
    reductions.add_argument(
        '--dpr', metavar='<role:count,role:count,...>', help="""Specify a
        deviation preserving reduction.""")
    reductions.add_argument(
        '--hr', metavar='<role:count,role:count,...>', help="""Specify a
        hierarchical reduction.""")


def parse_reduction(game, args):
    if args.dpr is not None:
        red_players = game.role_from_repr(args.dpr, dtype=int)
        red = reduction.deviation_preserving
    elif args.hr is not None:
        red_players = game.role_from_repr(args.hr, dtype=int)
        red = reduction.hierarchical
    else:
        red_players = None
        red = reduction.identity
    return red, red_players


types = {
    'zip': zipsched.create_scheduler,
    'sim': simsched.create_scheduler,
    'game': gamesched.create_scheduler,
    'eo': eosched.create_scheduler,
}


async def parse_scheduler(string):
    stype, args, *_ = itertools.chain(string.split(':', 1), [''])
    args = dict(s.split(':', 1) for s in args.split(',') if s)
    base = await types[stype](**args)
    if 'save' in args:
        base = SaveWrapper(base, args['save'])
    count = int(args.get('count', '1'))
    if count > 1:
        base = CountWrapper(base, count)
    return base


def add_scheduler_epilog(parser, help_indent=15):
    width, _ = shutil.get_terminal_size((80, 20))
    indent = ' ' * help_indent
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.description = textwrap.fill(
        ' '.join(parser.description.split()), width=width)

    epilog = ['scheduler specification:', '']
    epilog.append(textwrap.fill(
        ' '.join("""
A scheduler specification tells egta how to get samples of profiles. Their
format looks like `<type>:[param:value][,param:value]...`. Each scheduler type
and their supported parameters are listed below. For example:""".split()),
        width=width))
    epilog.extend(['', 'eo:game:827,mem:2048,time:600', ''])
    epilog.append(textwrap.fill(' '.join("""
Represents an egtaonline scheduler for game 827 with 2048 MB and 600 seconds
per observation.""".split())))
    epilog.append('')
    for name, func in itertools.chain(
            [('<global>', _global)], sorted(types.items())):
        start = '  ' + name
        desc = inspect.getdoc(func) or ''
        desc = ' '.join(desc.split())
        if len(start) < help_indent - 1:
            epilog.extend(textwrap.wrap(
                '{:{}}{}'.format(start, help_indent, desc), width=width,
                subsequent_indent=indent))
        else:
            epilog.append(start)
            epilog.extend(textwrap.wrap(
                desc, width=width, initial_indent=indent,
                subsequent_indent=indent))

        for pname, param in inspect.signature(func).parameters.items():
            if pname == '_':
                continue
            start = '    ' + pname
            desc = ''
            if param.annotation is not inspect._empty:
                desc += param.annotation
            if param.default is not None:
                desc += ' (default: {})'.format(param.default)
            desc = ' '.join(desc.split())
            if len(start) < help_indent - 1:
                epilog.extend(textwrap.wrap(
                    '{:{}}{}'.format(start, help_indent, desc), width=width,
                    subsequent_indent=indent))
            else:
                epilog.append(start)
                epilog.extend(textwrap.wrap(
                    desc, width=width, initial_indent=indent,
                    subsequent_indent=indent))
        epilog.append('')
    parser.epilog = '\n'.join(epilog)


def _global(
        count: """The number of samples to compute for one payoff
        observations."""='1',
        save: """A file to save all sampled profile data to as a sample
        game."""=None):
    """Global parameters that apply to all schedulers."""
    pass  # pragma: no cover


class SaveWrapper(savesched.SaveScheduler):
    def __init__(self, sched, dest):
        super().__init__(sched)
        self._dest = dest

    async def __aenter__(self):
        await self._sched.__aenter__()
        return self

    async def __aexit__(self, *args):
        with open(self._dest, 'w') as f:
            json.dump(self.get_game().to_json(), f)
        await self._sched.__aexit__(*args)


class CountWrapper(countsched.CountScheduler):
    async def __aenter__(self):
        await self._sched.__aenter__()
        return self

    async def __aexit__(self, *args):
        await self._sched.__aexit__(*args)



import abc
import asyncio

from gameanalysis import rsgame
from gameanalysis import mergegame


class AsyncGame(rsgame.GameLike):
    """An asynchronous game

    Supports asynchronous methods for ensuring particular payoff data"""

    @abc.abstractmethod
    async def get_restricted_game(self, rest):
        """Return a complete restricted game"""
        pass  # pragma: no cover

    @abc.abstractmethod
    async def get_deviation_game(self, rest, role_index=None):
        """Return a game with payoff data for all deviations

        If role index is specified, only deviations from that role are
        necessary."""
        pass  # pragma: no cover


class CompleteAsyncGame(AsyncGame):
    """A wrapper for a complete RsGame"""
    def __init__(self, game):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._game = game

    async def get_restricted_game(self, rest):
        return self._game.restrict(rest)

    async def get_deviation_game(self, rest, role_index=None):
        return self._game

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        return super().__eq__(other) and self._game == other._game

    def __str__(self):
        return repr(self._game)


def wrap(game):
    """Wrap a CompleteGame as an AsyncGame"""
    assert game.is_complete(), "must use a complete game"
    return CompleteAsyncGame(game)


class MergedAsyncGame(AsyncGame):
    """A lazy merging of two async games"""
    def __init__(self, agame0, agame1, t):
        super().__init__(
            agame0.role_names, agame0.strat_names, agame0.num_role_players)
        self._agame0 = agame0
        self._agame1 = agame1
        self._t = t

    async def get_restricted_game(self, rest):
        game0, game1 = await asyncio.gather(
            self._agame0.get_restricted_game(rest),
            self._agame1.get_restricted_game(rest))
        return mergegame.merge(game0, game1, self._t)

    async def get_deviation_game(self, rest, role_index=None):
        game0, game1 = await asyncio.gather(
            self._agame0.get_deviation_game(rest, role_index),
            self._agame1.get_deviation_game(rest, role_index))
        return mergegame.merge(game0, game1, self._t)

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        return (super().__eq__(other) and
                self._agame0 == other._agame0 and
                self._agame1 == other._agame1 and
                self._t == other._t)

    def __str__(self):
        return '{} {:g} - {:g} {}'.format(
            self._agame0, 1 - self._t, self._t, self._agame1)


def merge(agame0, agame1, t):
    """Merge two async games"""
    assert rsgame.emptygame_copy(agame0) == rsgame.emptygame_copy(agame1)
    return MergedAsyncGame(agame0, agame1, t)



"""An abstract profile scheduler"""
import abc

from gameanalysis import rsgame


class Scheduler(rsgame.GameLike):
    """A profile scheduler

    It must be a context manager for its resources."""

    @abc.abstractmethod
    async def sample_payoffs(self, profile):
        """Schedule a profile

        Return a future for the payoff data"""
        pass  # pragma: no cover



import sys

from gameanalysis import paygame
from gameanalysis import gamereader

from egta import gamesched


async def create_scheduler(
        game: "A file with the game data. `-` is interpreted as stdin."='-',
        sample: """Treat the game as a sample game. This doesn't require a value,
        just a trialing colon."""=None, **_):
    """A scheduler that samples payoff data from an existing game. If sample is
    specified, payoffs will be a random payoff from each payoff for the
    profile."""
    if game == '-':
        rsgame = gamereader.load(sys.stdin)
    else:
        with open(game) as f:
            rsgame = gamereader.load(f)

    if sample is not None:
        return Sgs(paygame.samplegame_copy(rsgame))
    else:
        return Rgs(rsgame)


class Sgs(gamesched.SampleGameScheduler):
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass


class Rgs(gamesched.RsGameScheduler):
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass



"""Script utility for running inner loop"""
import asyncio
import json
import logging

from gameanalysis import regret

from egta import innerloop
from egta import schedgame
from egta.script import utils


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'quiesce', help="""Compute equilibria using the quiesce procedure""",
        description="""Samples profiles from small restricted strategy sets,
        expanding set support by best responses to candidate restricted game
        equilibria. For games with a large number of players, a reduction
        should be specified. The result is a list where each element specifies
        an "equilibrium".""")
    parser.add_argument(
        'scheduler', metavar='<sched-spec>', help="""A scheduler specification,
        see below.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=0.1,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    parser.add_argument(
        '--max-restrict-size', metavar='<support>', type=int, default=3,
        help="""Support size threshold, beyond which restricted games are not
        required to be explored.  (default: %(default)d)""")
    parser.add_argument(
        '--num-equilibria', metavar='<num>', type=int, default=1,
        help="""Number of equilibria requested to be found. This is mainly
        useful when game contains known degenerate equilibria, but those
        strategies are still useful as deviating strategies. (default:
        %(default)d)""")
    parser.add_argument(
        '--num-backups', metavar='<num>', type=int, default=1, help="""Number
        of backup restricted strategy set to pop at a time, when no equilibria
        are confirmed in initial required set.  When games get to this point
        they can quiesce slowly because this by default pops one at a time.
        Increasing this number can get games like tis to quiesce more quickly,
        but naturally, also schedules more, potentially unnecessary,
        simulations. (default: %(default)d)""")
    parser.add_argument(
        '--dev-by-role', action='store_true', help="""Explore deviations in
        role order instead of all at once. By default, when checking for
        beneficial deviations, all role deviations are scheduled at the same
        time. Setting this will check one role at a time. If a beneficial
        deviation is found, then that restricted strategy set is scheduled
        without exploring deviations from the other roles.""")
    parser.add_argument(
        '--one', action='store_true', help="""Guarantee that an equilibrium is
        found in every restricted game. This may take up to exponential
        time.""")
    utils.add_reductions(parser)
    utils.add_scheduler_epilog(parser)
    parser.run = run
    return parser


async def run(args):
    sched = await utils.parse_scheduler(args.scheduler)
    red, red_players = utils.parse_reduction(sched, args)
    agame = schedgame.schedgame(sched, red, red_players)

    async def get_regret(eqm):
        game = await agame.get_deviation_game(eqm > 0)
        return float(regret.mixture_regret(game, eqm))

    async with sched:
        eqa = await innerloop.inner_loop(
            agame, regret_thresh=args.regret_thresh,
            dist_thresh=args.dist_thresh,
            restricted_game_size=args.max_restrict_size,
            num_equilibria=args.num_equilibria, num_backups=args.num_backups,
            devs_by_role=args.dev_by_role, at_least_one=args.one)
        regrets = await asyncio.gather(*[
            get_regret(eqm) for eqm in eqa])

    logging.error(
        "quiesce finished finding %d equilibria:\n%s",
        eqa.shape[0], '\n'.join(
            '{:d}) {} with regret {:g}'.format(
                i, sched.mixture_to_repr(eqm), reg)
            for i, (eqm, reg) in enumerate(zip(eqa, regrets), 1)))

    json.dump([{'equilibrium': sched.mixture_to_json(eqm), 'regret': reg}
               for eqm, reg in zip(eqa, regrets)],
              args.output)
    args.output.write('\n')



"""A scheduler that gets payoffs from a local simulation"""
import asyncio
import collections
import contextlib
import json
import logging
import subprocess

from gameanalysis import paygame
from gameanalysis import rsgame

from egta import profsched


class SimulationScheduler(profsched.Scheduler):
    """Schedule profiles using a command line program

    Parameters
    ----------
    game : RsGame
        A gameanalysis game that indicates how array profiles should be turned
        into json profiles.
    config : {key: value}
        A dictionary mapping string keys to values that will be passed to the
        simulator in the standard simulation spec format.
    command : [str]
        A list of strings that represents a command line program to run. This
        program must accept simulation spec files as flushed lines of input to
        standard in, and write the resulting output as an observation to
        standard out. After all input lines have been read, this must flush the
        output otherwise this could hang waiting for results that are trapped
        in a buffer.
    buff_size : int, optional
        The maximum number of bytes to send to the command at a time. The
        default should be fine for most applications, but if you know your
        machine has a larger or smaller buffer size, setting this accurately
        will prevent unnecessary blocking.
    """

    def __init__(self, game, config, command, buff_size=4096):
        super().__init__(
            game.role_names, game.strat_names, game.num_role_players)
        self._game = paygame.game_copy(rsgame.emptygame_copy(game))
        self._base = {'configuration': config}
        self.command = command
        self.buff_size = buff_size

        self._is_open = False
        self._proc = None
        self._reader = None
        self._read_queue = asyncio.Queue()
        self._write_lock = asyncio.Lock()
        self._buffer_empty = asyncio.Event()
        self._buffer_bytes = 0
        self._line_bytes = collections.deque()

        self._buffer_empty.set()

    async def sample_payoffs(self, profile):
        assert self._is_open, "not open"

        self._base['assignment'] = self._game.profile_to_json(profile)
        bprof = json.dumps(self._base, separators=(',', ':')).encode('utf8')
        size = len(bprof) + 1
        assert size < self.buff_size, \
            "profile could not be written to buffer without blocking"
        async with self._write_lock:
            self._buffer_bytes += size
            self._line_bytes.appendleft(size)
            if self._buffer_bytes >= self.buff_size:
                self._buffer_empty.clear()
                await self._buffer_empty.wait()

            got_data = asyncio.Event()
            line = [None]
            self._read_queue.put_nowait((line, got_data))

            self._proc.stdin.write(bprof)
            self._proc.stdin.write(b'\n')
            try:
                await self._proc.stdin.drain()
            except ConnectionError:  # pragma: no cover race condition
                raise RuntimeError("process died unexpectedly")

        logging.debug("scheduled profile: %s",
                      self._game.profile_to_repr(profile))
        await got_data.wait()
        if self._reader.done() and self._reader.exception() is not None:
            raise self._reader.exception()
        jpays = json.loads(line[0].decode('utf8'))
        payoffs = self._game.payoff_from_json(jpays)
        payoffs.setflags(write=False)
        logging.debug("read payoff for profile: %s",
                      self.profile_to_repr(profile))
        return payoffs

    async def _read(self):
        while True:
            line, got_data = await self._read_queue.get()
            try:
                line[0] = await self._proc.stdout.readline()
                if not len(line[0]):
                    raise RuntimeError("process died unexpectedly")
                self._buffer_bytes -= self._line_bytes.pop()
                if self._buffer_bytes < self.buff_size:  # pragma: no branch
                    self._buffer_empty.set()
            finally:
                got_data.set()

    async def open(self):
        assert not self._is_open, "can't open twice"
        assert self._proc is None
        assert self._reader is None
        try:
            self._proc = await asyncio.create_subprocess_exec(
                *self.command, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
            self._reader = asyncio.ensure_future(self._read())
            self._is_open = True
        except Exception as ex:
            await self.close()
            raise ex
        return self

    async def close(self):
        self._is_open = False

        if self._reader is not None:
            self._reader.cancel()
            with contextlib.suppress(Exception):
                await self._reader
            self._reader = None

        if self._proc is not None:
            with contextlib.suppress(ProcessLookupError):
                self._proc.terminate()
            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self._proc.wait(), 0.25)
            with contextlib.suppress(ProcessLookupError):
                self._proc.kill()
            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self._proc.wait(), 0.25)
            self._proc = None

        while not self._read_queue.empty():
            self._read_queue.get_nowait()
        self._buffer_empty.set()
        self._buffer_bytes = 0
        self._line_bytes.clear()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, *args):
        await self.close()

    def __str__(self):
        return ' '.join(self.command)


def simsched(game, config, command, buff_size=4096):
    return SimulationScheduler(game, config, command, buff_size=buff_size)



import asyncio

import numpy as np
import numpy.random as rand
import pytest
from gameanalysis import agggen
from gameanalysis import gamegen
from gameanalysis import rsgame

from egta import gamesched


@pytest.mark.asyncio
async def test_basic_profile():
    game = gamegen.game([4, 3], [3, 4])
    profs = game.random_profiles(20)

    sched = gamesched.gamesched(game)
    assert (rsgame.emptygame_copy(sched) ==
            rsgame.emptygame_copy(game))
    paylist = await asyncio.gather(*[
        sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)


@pytest.mark.asyncio
async def test_basic_profile_sample():
    sgame = gamegen.samplegame([4, 3], [3, 4])
    profs = sgame.random_profiles(20)

    sched = gamesched.samplegamesched(sgame)
    assert (rsgame.emptygame_copy(sched) ==
            rsgame.emptygame_copy(sgame))
    paylist = await asyncio.gather(*[
        sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)


@pytest.mark.asyncio
async def test_duplicate_profile_sample():
    sgame = gamegen.samplegame([4, 3], [3, 4], 0)
    profs = sgame.random_profiles(20)

    sched = gamesched.samplegamesched(sgame)
    paylist1 = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays1 = np.stack(paylist1)
    paylist2 = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays2 = np.stack(paylist2)
    assert np.allclose(pays1[profs == 0], 0)
    assert np.allclose(pays2[profs == 0], 0)
    assert np.allclose(pays1, pays2)


@pytest.mark.asyncio
async def test_basic_profile_aggfn():
    agame = agggen.normal_aggfn([4, 3], [3, 4], 5)
    profs = agame.random_profiles(20)

    sched = gamesched.gamesched(agame)
    paylist = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)


@pytest.mark.asyncio
async def test_noise_profile():
    sgame = gamegen.samplegame([4, 3], [3, 4])
    profs = sgame.random_profiles(20)

    sched = gamesched.samplegamesched(
        sgame, lambda w: rand.normal(0, w, sgame.num_strats),
        lambda: (rand.random(),))
    paylist = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays = np.stack(paylist)
    assert np.allclose(pays[profs == 0], 0)


@pytest.mark.asyncio
async def test_duplicate_prof():
    game = gamegen.game([4, 3], [3, 4])
    profs = game.random_profiles(20)

    sched = gamesched.gamesched(game)
    paylist1 = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays1 = np.stack(paylist1)
    paylist2 = await asyncio.gather(*[sched.sample_payoffs(p) for p in profs])
    pays2 = np.stack(paylist2)
    assert np.allclose(pays1[profs == 0], 0)
    assert np.allclose(pays2[profs == 0], 0)
    assert np.allclose(pays1, pays2)



import json
import sys

from gameanalysis import gamereader

from egta import zipsched


async def create_scheduler(
        game: """A file with the description of the game to generate profiles
        from. Only the basic game structure is necessary. `-` is interpreted as
        stdin."""='-',
        procs: """The maximum number of processes used to generate payoff
        data."""='4',
        zipf: """The zipfile to run. This must be identical to what would be
        uploaded to egtaonline. (required)"""=None,
        conf: """A file with the specific configuration to load. `-` is
        interpreted as stdin. (default: {})"""=None,
        count: """This scheduler is more efficient at sampling several
        profiles."""='1', **_):
    """Get payoffs from an egtaonline style simulator. This won't be as
    efficient as using `sim`, but requires no modifications from an existing
    egtaonline scheduler."""
    assert zipf is not None, '"zipf" must be specified'
    max_procs = int(procs)
    count = int(count)

    if game == '-':
        rsgame = gamereader.load(sys.stdin)
    else:
        with open(game) as f:
            rsgame = gamereader.load(f)

    if conf is None:
        config = {}
    elif conf == '-':
        config = json.load(sys.stdin)
    else:
        with open(conf) as f:
            config = json.load(f)

    return ZipSched(
        rsgame, config, zipf, max_procs=max_procs, simultaneous_obs=count)


class ZipSched(zipsched.ZipScheduler):
    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, *args):
        return self.__exit__(*args)



import pytest
import numpy as np
from gameanalysis import gamegen

from egta import gamesched
from egta import schedgame


sizes = [
    ([3], [3]),
    ([3, 2], [2, 3]),
    ([2, 2, 2], [2, 2, 2]),
]


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
@pytest.mark.parametrize('_', range(20))
async def test_random_caching(players, strats, _):
    game = gamegen.samplegame(players, strats)
    rest1, rest2 = game.random_restrictions(2)
    mixes = np.random.random((2, game.num_strats))
    mixes *= [rest1, rest2]
    mixes /= np.add.reduceat(mixes, game.role_starts, 1).repeat(
        game.num_role_strats, 1)
    mix1, mix2 = mixes
    sched = gamesched.SampleGameScheduler(game)
    sgame = schedgame.schedgame(sched)

    rgame11 = await sgame.get_restricted_game(rest1)
    rgame21 = await sgame.get_restricted_game(rest2)
    devs11 = await sgame.get_deviation_game(rest1)
    devs21 = await sgame.get_deviation_game(rest2)

    rgame12 = await sgame.get_restricted_game(rest1)
    rgame22 = await sgame.get_restricted_game(rest2)
    assert rgame11 == rgame12
    assert rgame21 == rgame22

    devs12 = await sgame.get_deviation_game(rest1)
    devs22 = await sgame.get_deviation_game(rest2)
    assert devs11 == devs12
    assert devs21 == devs22


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
@pytest.mark.parametrize('_', range(20))
async def test_random_complete_dev(players, strats, _):
    game = gamegen.samplegame(players, strats)
    sched = gamesched.SampleGameScheduler(game)
    sgame = schedgame.schedgame(sched)
    mix = sgame.random_sparse_mixture()
    supp = mix > 0
    dev_game = await sgame.get_deviation_game(supp)
    devs, jac = dev_game.deviation_payoffs(mix, jacobian=True)
    assert not np.isnan(devs).any()
    assert not np.isnan(jac[supp]).any()
    assert np.isnan(jac[~supp]).all()
    for r in range(sgame.num_roles):
        mask = r == sgame.role_indices
        dev_game = await sgame.get_deviation_game(supp, role_index=r)
        rdevs = dev_game.deviation_payoffs(mix)
        assert np.allclose(rdevs[supp], devs[supp])
        assert np.allclose(rdevs[mask], devs[mask])
        assert supp[~mask].all() or np.isnan(rdevs[~mask]).any()


@pytest.mark.asyncio
@pytest.mark.parametrize('players,strats', sizes)
@pytest.mark.parametrize('_', range(20))
async def test_random_normalize(players, strats, _):
    game = gamegen.samplegame(players, strats)
    sched = gamesched.SampleGameScheduler(game)
    sgame = schedgame.schedgame(sched)
    rgame = await sgame.get_restricted_game(np.ones(game.num_strats, bool))
    ngame = rgame.normalize()
    assert np.all(ngame.payoffs() >= -1e-7)
    assert np.all(ngame.payoffs() <= 1 + 1e-7)



import json
import sys

from gameanalysis import gamereader

from egta import simsched


async def create_scheduler(
        game: """A file with the description of the game to generate profiles
        from. Only the basic game structure is necessary. `-` is interpreted as
        stdin."""='-',
        conf: """A file with the specific configuration to load. `-` is
        interpreted as stdin. (default: {})"""=None,
        command: """The command to run. This is space delimited, heavily
        restricting what can actually be passed. Currently the best work around
        is writing a simple shell wrapper and then executing `bash
        wrapper.sh`. (required)"""=None,
        buff: """Maximum line buffer to prevent deadlock with the subprocess.
        This default is fine unless you know what you're doing."""='4096',
        **_):
    """Get payoffs from a command line simulator. The simulator will get passed
    a new compressed simulation spec file on each line of stdin, and is
    expected to write a standard observation file to each line of stdout in the
    same order."""
    assert command is not None, '"command" must be specified'
    buff_size = int(buff)

    if game == '-':
        rsgame = gamereader.load(sys.stdin)
    else:
        with open(game) as f:
            rsgame = gamereader.load(f)

    if conf is None:
        config = {}
    elif conf == '-':
        config = json.load(sys.stdin)
    else:
        with open(conf) as f:
            config = json.load(f)

    return simsched.simsched(
        rsgame, config, command.split(), buff_size=buff_size)
